# -*- coding: utf-8 -*-
"""Utility functions for reference handling, coordinate flips, and file operations."""

import logging
import os

import nibabel as nib
from nibabel.streamlines.array_sequence import ArraySequence
from nibabel.streamlines.tractogram import Tractogram, TractogramItem
import numpy as np

try:
    import dipy

    dipy_available = True
except ImportError:
    dipy_available = False


def close_or_delete_mmap(obj):
    """Close the memory-mapped file if it exists, otherwise set the object to None.

    Parameters
    ----------
    obj : object
        The object that potentially has a memory-mapped file to be closed.
    """
    if hasattr(obj, "_mmap") and obj._mmap is not None:
        obj._mmap.close()
    elif isinstance(obj, ArraySequence):
        close_or_delete_mmap(obj._data)
        close_or_delete_mmap(obj._offsets)
        close_or_delete_mmap(obj._lengths)
    elif isinstance(obj, np.memmap):
        del obj
    else:
        logging.debug("Object to be close or deleted must be np.memmap")


def split_name_with_gz(filename):
    """Return the clean basename and extension of a file.

    Correctly manages the ".nii.gz" extensions.

    Parameters
    ----------
    filename : str
        The filename to clean.

    Returns
    -------
    base : str
        Clean basename.
    ext : str
        The full extension.
    """
    base, ext = os.path.splitext(filename)

    if ext == ".gz":
        # Test if we have a .nii additional extension
        temp_base, add_ext = os.path.splitext(base)

        if add_ext == ".nii" or add_ext == ".trk":
            ext = add_ext + ext
            base = temp_base

    return base, ext


def get_reference_info_wrapper(reference):  # noqa: C901
    """Extract spatial attributes from a reference object.

    Parameters
    ----------
    reference : str or dict or Nifti1Image or TrkFile or Nifti1Header or TrxFile
        Reference that provides the spatial attribute.

    Returns
    -------
    affine : ndarray (4, 4)
        Transformation of VOX to RASMM, np.float32.
    dimensions : ndarray (3,)
        Volume shape for each axis, int16.
    voxel_sizes : ndarray (3,)
        Size of voxel for each axis, float32.
    voxel_order : str
        Typically 'RAS' or 'LPS'.
    """
    from trx import trx_file_memmap

    is_nifti = False
    is_trk = False
    is_sft = False
    is_trx = False
    if isinstance(reference, str):
        _, ext = split_name_with_gz(reference)
        if ext in [".nii", ".nii.gz"]:
            header = nib.load(reference).header
            is_nifti = True
        elif ext == ".trk":
            header = nib.streamlines.load(reference, lazy_load=True).header
            is_trk = True
        elif ext == ".trx":
            header = trx_file_memmap.load(reference).header
            is_trx = True
    elif isinstance(reference, trx_file_memmap.TrxFile):
        header = reference.header
        is_trx = True
    elif isinstance(reference, nib.nifti1.Nifti1Image):
        header = reference.header
        is_nifti = True
    elif isinstance(reference, nib.streamlines.trk.TrkFile):
        header = reference.header
        is_trk = True
    elif isinstance(reference, nib.nifti1.Nifti1Header):
        header = reference
        is_nifti = True
    elif isinstance(reference, dict) and "magic_number" in reference:
        header = reference
        is_trk = True
    elif isinstance(reference, dict) and "NB_VERTICES" in reference:
        header = reference
        is_trx = True
    elif dipy_available and isinstance(
        reference, dipy.io.stateful_tractogram.StatefulTractogram
    ):
        is_sft = True

    if is_nifti:
        affine = header.get_best_affine()
        dimensions = header["dim"][1:4]
        voxel_sizes = header["pixdim"][1:4]

        if not affine[0:3, 0:3].any():
            raise ValueError(
                "Invalid affine, contains only zeros."
                "Cannot determine voxel order from transformation"
            )
        voxel_order = "".join(nib.aff2axcodes(affine))
    elif is_trk:
        affine = header["voxel_to_rasmm"]
        dimensions = header["dimensions"]
        voxel_sizes = header["voxel_sizes"]
        voxel_order = header["voxel_order"]
    elif is_sft:
        affine, dimensions, voxel_sizes, voxel_order = reference.space_attributes
    elif is_trx:
        affine = header["VOXEL_TO_RASMM"]
        dimensions = header["DIMENSIONS"]
        voxel_sizes = nib.affines.voxel_sizes(affine)
        voxel_order = "".join(nib.aff2axcodes(affine))
    else:
        raise TypeError("Input reference is not one of the supported format")

    if isinstance(voxel_order, np.bytes_):
        voxel_order = voxel_order.decode("utf-8")

    if dipy_available:
        from dipy.io.utils import is_reference_info_valid

        is_reference_info_valid(affine, dimensions, voxel_sizes, voxel_order)

    return affine, dimensions, voxel_sizes, voxel_order


def is_header_compatible(reference_1, reference_2):
    """Compare the spatial attributes of 2 references.

    Parameters
    ----------
    reference_1 : Nifti or Trk filename, Nifti1Image or TrkFile,
        Nifti1Header or trk.header (dict)
        Reference that provides the spatial attribute.
    reference_2 : Nifti or Trk filename, Nifti1Image or TrkFile,
        Nifti1Header or trk.header (dict)
        Reference that provides the spatial attribute.

    Returns
    -------
    bool
        Whether all the spatial attributes match.
    """

    affine_1, dimensions_1, voxel_sizes_1, voxel_order_1 = get_reference_info_wrapper(
        reference_1
    )
    affine_2, dimensions_2, voxel_sizes_2, voxel_order_2 = get_reference_info_wrapper(
        reference_2
    )

    identical_header = True
    if not np.allclose(affine_1, affine_2, rtol=1e-03, atol=1e-03):
        logging.error("Affine not equal")
        identical_header = False

    if not np.array_equal(dimensions_1, dimensions_2):
        logging.error("Dimensions not equal")
        identical_header = False

    if not np.allclose(voxel_sizes_1, voxel_sizes_2, rtol=1e-03, atol=1e-03):
        logging.error("Voxel_size not equal")
        identical_header = False

    if voxel_order_1 != voxel_order_2:
        logging.error("Voxel_order not equal")
        identical_header = False

    return identical_header


def get_axis_shift_vector(flip_axes):
    """Return a shift vector for the given axes.

    Parameters
    ----------
    flip_axes : list of str
        String containing the axis to flip.
        Possible values are 'x', 'y', 'z'.

    Returns
    -------
    shift_vector : np.ndarray (3,)
        Vector containing the axis to shift.
        Possible values are -1, 0.
    """
    shift_vector = np.zeros(3)
    if "x" in flip_axes:
        shift_vector[0] = -1.0
    if "y" in flip_axes:
        shift_vector[1] = -1.0
    if "z" in flip_axes:
        shift_vector[2] = -1.0

    return shift_vector


def get_axis_flip_vector(flip_axes):
    """Return a flip vector for the given axes.

    Parameters
    ----------
    flip_axes : list of str
        String containing the axis to flip.
        Possible values are 'x', 'y', 'z'.

    Returns
    -------
    flip_vector : np.ndarray (3,)
        Vector containing the axis to flip.
        Possible values are -1, 1.
    """
    flip_vector = np.ones(3)
    if "x" in flip_axes:
        flip_vector[0] = -1.0
    if "y" in flip_axes:
        flip_vector[1] = -1.0
    if "z" in flip_axes:
        flip_vector[2] = -1.0

    return flip_vector


def get_shift_vector(sft):
    """Return the shift vector for flipping a tractogram.

    When flipping a tractogram the shift vector is used to change the origin
    of the grid from the corner to the center of the grid.

    Parameters
    ----------
    sft : StatefulTractogram
        StatefulTractogram object.

    Returns
    -------
    shift_vector : ndarray
        Shift vector to apply to the streamlines.
    """
    dims = sft.space_attributes[1]
    shift_vector = -1.0 * (np.array(dims) / 2.0)

    return shift_vector


def flip_sft(sft, flip_axes):
    """Flip the streamlines in a StatefulTractogram.

    Use the spatial information to flip according to the center
    of the grid.

    Parameters
    ----------
    sft : StatefulTractogram
        StatefulTractogram to flip.
    flip_axes : list of str
        Axes to flip.
        Possible values are 'x', 'y', 'z'.

    Returns
    -------
    sft : StatefulTractogram
        StatefulTractogram with flipped axes.
    """
    if not dipy_available:
        logging.error(
            "Dipy library is missing, cannot use functions related "
            "to the StatefulTractogram."
        )
        return None

    flip_vector = get_axis_flip_vector(flip_axes)
    shift_vector = get_shift_vector(sft)

    flipped_streamlines = []
    for streamline in sft.streamlines:
        mod_streamline = streamline + shift_vector
        mod_streamline *= flip_vector
        mod_streamline -= shift_vector
        flipped_streamlines.append(mod_streamline)

    from dipy.io.stateful_tractogram import StatefulTractogram

    new_sft = StatefulTractogram.from_sft(
        flipped_streamlines,
        sft,
        data_per_point=sft.data_per_point,
        data_per_streamline=sft.data_per_streamline,
    )
    return new_sft


def load_matrix_in_any_format(filepath):
    """Load a matrix from a txt file OR a npy file.

    Parameters
    ----------
    filepath : str
        Path to the matrix file.

    Returns
    -------
    matrix : numpy.ndarray
        The matrix.
    """
    _, ext = os.path.splitext(filepath)
    if ext == ".txt":
        data = np.loadtxt(filepath)
    elif ext == ".npy":
        data = np.load(filepath)
    else:
        raise ValueError("Extension {} is not supported".format(ext))

    return data


def get_reverse_enum(space_str, origin_str):
    """Convert string representation to enums for the StatefulTractogram.

    Parameters
    ----------
    space_str : str
        String representing the space.
    origin_str : str
        String representing the origin.

    Returns
    -------
    space : Space
        Space enum value.
    origin : Origin
        Origin enum value.
    """
    if not dipy_available:
        logging.error(
            "Dipy library is missing, cannot use functions related "
            "to the StatefulTractogram."
        )
        return None
    from dipy.io.stateful_tractogram import Origin, Space

    origin = Origin.NIFTI if origin_str.lower() == "nifti" else Origin.TRACKVIS
    if space_str.lower() == "rasmm":
        space = Space.RASMM
    elif space_str.lower() == "voxmm":
        space = Space.VOXMM
    else:
        space = Space.VOX

    return space, origin


def convert_data_dict_to_tractogram(data):
    """Convert data from a lazy tractogram to a tractogram.

    Parameters
    ----------
    data : dict
        The data dictionary to convert into a nibabel tractogram.

    Returns
    -------
    Tractogram
        A Tractogram object.
    """
    streamlines = ArraySequence(data["strs"])
    streamlines._data = streamlines._data

    for key in data["dps"]:
        shape = (len(streamlines), len(data["dps"][key]) // len(streamlines))
        data["dps"][key] = np.array(data["dps"][key]).reshape(shape)

    for key in data["dpv"]:
        shape = (
            len(streamlines._data),
            len(data["dpv"][key]) // len(streamlines._data),
        )
        data["dpv"][key] = np.array(data["dpv"][key]).reshape(shape)

        tmp_arr = ArraySequence()
        tmp_arr._data = data["dpv"][key]
        tmp_arr._offsets = streamlines._offsets
        tmp_arr._lengths = streamlines._lengths
        data["dpv"][key] = tmp_arr

    obj = Tractogram(
        streamlines, data_per_point=data["dpv"], data_per_streamline=data["dps"]
    )

    return obj


def append_generator_to_dict(gen, data):
    """Append items yielded by a tractogram generator into data dict.

    Parameters
    ----------
    gen : TractogramItem class instance or np.ndarray
        Item produced by a tractogram generator. Structured entries include
        per-point and per-streamline metadata.
    data : dict
        Accumulator containing ``strs`` (positions), ``dpv`` and ``dps``
        dictionaries that will be extended in-place.

    Returns
    -------
    None
        The function mutates ``data`` and returns ``None``.
    """

    if isinstance(gen, TractogramItem):
        data["strs"].append(gen.streamline.tolist())
        for key in gen.data_for_points:
            if key not in data["dpv"]:
                data["dpv"][key] = np.array([])
            data["dpv"][key] = np.append(data["dpv"][key], gen.data_for_points[key])
        for key in gen.data_for_streamline:
            if key not in data["dps"]:
                data["dps"][key] = np.array([])
            data["dps"][key] = np.append(data["dps"][key], gen.data_for_streamline[key])
    else:
        data["strs"].append(gen.tolist())


def verify_trx_dtype(trx, dict_dtype):  # noqa: C901
    """Verify that data dtypes in the trx match the given dict.

    Parameters
    ----------
    trx : Tractogram
        Tractogram to verify.
    dict_dtype : dict
        Dictionary containing all elements dtype to verify.

    Returns
    -------
    bool
        True if the dtype is the same, False otherwise.
    """
    identical = True
    for key in dict_dtype:
        if key == "positions":
            if trx.streamlines._data.dtype != dict_dtype[key]:
                logging.warning("Positions dtype is different")
                identical = False
        elif key == "offsets":
            if trx.streamlines._offsets.dtype != dict_dtype[key]:
                logging.warning("Offsets dtype is different")
                identical = False
        elif key == "dpv":
            for key_dpv in dict_dtype[key]:
                if trx.data_per_vertex[key_dpv]._data.dtype != dict_dtype[key][key_dpv]:
                    logging.warning(
                        "Data per vertex ({}) dtype is different".format(key_dpv)
                    )
                    identical = False
        elif key == "dps":
            for key_dps in dict_dtype[key]:
                if trx.data_per_streamline[key_dps].dtype != dict_dtype[key][key_dps]:
                    logging.warning(
                        "Data per streamline ({}) dtype is different".format(key_dps)
                    )
                    identical = False
        elif key == "dpg":
            for key_group in dict_dtype[key]:
                for key_dpg in dict_dtype[key][key_group]:
                    if (
                        trx.data_per_point[key_group][key_dpg].dtype
                        != dict_dtype[key][key_group][key_dpg]
                    ):
                        logging.warning(
                            "Data per group ({}) dtype is different".format(key_dpg)
                        )
                        identical = False
        elif key == "groups":
            for key_group in dict_dtype[key]:
                if (
                    trx.data_per_point[key_group]._data.dtype
                    != dict_dtype[key][key_group]
                ):
                    logging.warning(
                        "Data per group ({}) dtype is different".format(key_group)
                    )
                    identical = False

    return identical



# -*- coding: utf-8 -*-
"""Tests for CLI commands and workflow functions."""

import os
import tempfile

from deepdiff import DeepDiff
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal, assert_equal
import pytest

try:
    from dipy.io.streamline import load_tractogram

    dipy_available = True
except ImportError:
    dipy_available = False

from trx.fetcher import fetch_data, get_home, get_testing_files_dict
import trx.trx_file_memmap as tmm
from trx.workflows import (
    convert_dsi_studio,
    convert_tractogram,
    generate_trx_from_scratch,
    manipulate_trx_datatype,
    validate_tractogram,
)

# If they already exist, this only takes 5 seconds (check md5sum)
fetch_data(
    get_testing_files_dict(),
    keys=["DSI.zip", "trx_from_scratch.zip", "gold_standard.zip"],
)


def _normalize_dtype_dict(dtype_dict):
    """Normalize dtype dict to use explicit little-endian byte order.

    On little-endian systems, numpy may use '=' (native) or '<' (explicit)
    interchangeably. This normalizes all dtypes to '<' for consistent comparison.
    """
    normalized = {}
    for key, value in dtype_dict.items():
        if isinstance(value, dict):
            normalized[key] = _normalize_dtype_dict(value)
        elif isinstance(value, np.dtype):
            # Normalize to little-endian for multi-byte types
            if value.byteorder == "=" and value.itemsize > 1:
                normalized[key] = value.newbyteorder("<")
            else:
                normalized[key] = value
        else:
            normalized[key] = value
    return normalized


# Tests for standalone CLI commands (trx_* commands)
class TestStandaloneCommands:
    """Tests for standalone CLI commands."""

    def test_help_option_convert_dsi(self, script_runner):
        ret = script_runner.run(["trx_convert_dsi_studio", "--help"])
        assert ret.success

    def test_help_option_convert(self, script_runner):
        ret = script_runner.run(["trx_convert_tractogram", "--help"])
        assert ret.success

    def test_help_option_generate_from_scratch(self, script_runner):
        ret = script_runner.run(["trx_generate_from_scratch", "--help"])
        assert ret.success

    def test_help_option_concatenate(self, script_runner):
        ret = script_runner.run(["trx_concatenate_tractograms", "--help"])
        assert ret.success

    def test_help_option_manipulate(self, script_runner):
        ret = script_runner.run(["trx_manipulate_datatype", "--help"])
        assert ret.success

    def test_help_option_compare(self, script_runner):
        ret = script_runner.run(["trx_simple_compare", "--help"])
        assert ret.success

    def test_help_option_validate(self, script_runner):
        ret = script_runner.run(["trx_validate", "--help"])
        assert ret.success

    def test_help_option_verify_header(self, script_runner):
        ret = script_runner.run(["trx_verify_header_compatibility", "--help"])
        assert ret.success

    def test_help_option_visualize(self, script_runner):
        ret = script_runner.run(["trx_visualize_overlap", "--help"])
        assert ret.success

    def test_help_option_info(self, script_runner):
        ret = script_runner.run(["trx_info", "--help"])
        assert ret.success


# Tests for unified trx CLI
class TestUnifiedCLI:
    """Tests for the unified trx CLI."""

    def test_trx_help(self, script_runner):
        ret = script_runner.run(["trx", "--help"])
        assert ret.success

    def test_trx_concatenate_help(self, script_runner):
        ret = script_runner.run(["trx", "concatenate", "--help"])
        assert ret.success

    def test_trx_convert_help(self, script_runner):
        ret = script_runner.run(["trx", "convert", "--help"])
        assert ret.success

    def test_trx_convert_dsi_help(self, script_runner):
        ret = script_runner.run(["trx", "convert-dsi", "--help"])
        assert ret.success

    def test_trx_generate_help(self, script_runner):
        ret = script_runner.run(["trx", "generate", "--help"])
        assert ret.success

    def test_trx_manipulate_dtype_help(self, script_runner):
        ret = script_runner.run(["trx", "manipulate-dtype", "--help"])
        assert ret.success

    def test_trx_compare_help(self, script_runner):
        ret = script_runner.run(["trx", "compare", "--help"])
        assert ret.success

    def test_trx_validate_help(self, script_runner):
        ret = script_runner.run(["trx", "validate", "--help"])
        assert ret.success

    def test_trx_verify_header_help(self, script_runner):
        ret = script_runner.run(["trx", "verify-header", "--help"])
        assert ret.success

    def test_trx_visualize_help(self, script_runner):
        ret = script_runner.run(["trx", "visualize", "--help"])
        assert ret.success

    def test_trx_info_help(self, script_runner):
        ret = script_runner.run(["trx", "info", "--help"])
        assert ret.success

    def test_trx_info_execution(self, script_runner):
        """Test trx info command execution on a real TRX file."""
        trx_path = os.path.join(get_home(), "gold_standard", "gs.trx")
        ret = script_runner.run(["trx", "info", trx_path])
        assert ret.success
        # Check key output elements
        assert "VOXEL_TO_RASMM" in ret.stdout
        assert "DIMENSIONS" in ret.stdout
        assert "streamline_count" in ret.stdout
        assert "vertex_count" in ret.stdout
        assert "Archive contents:" in ret.stdout

    def test_trx_info_wrong_extension(self, script_runner):
        """Test trx info rejects non-TRX files."""
        tck_path = os.path.join(get_home(), "gold_standard", "gs.tck")
        ret = script_runner.run(["trx", "info", tck_path])
        assert not ret.success
        assert "not a TRX file" in ret.stderr

    def test_trx_info_file_not_found(self, script_runner):
        """Test trx info handles missing files."""
        ret = script_runner.run(["trx", "info", "nonexistent.trx"])
        assert not ret.success
        assert "does not exist" in ret.stderr


# Tests for workflow functions
class TestWorkflowFunctions:
    """Tests for workflow functions."""

    @pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
    def test_execution_convert_dsi(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            in_trk = os.path.join(get_home(), "DSI", "CC.trk.gz")
            in_nii = os.path.join(get_home(), "DSI", "CC.nii.gz")
            exp_data = os.path.join(get_home(), "DSI", "CC_fix_data.npy")
            exp_offsets = os.path.join(get_home(), "DSI", "CC_fix_offsets.npy")
            out_fix_path = os.path.join(tmp_dir, "fixed.trk")
            convert_dsi_studio(
                in_trk, in_nii, out_fix_path, remove_invalid=False, keep_invalid=True
            )

            data_fix = np.load(exp_data)
            offsets_fix = np.load(exp_offsets)

            sft = load_tractogram(out_fix_path, "same")
            assert_equal(sft.streamlines._data, data_fix)
            assert_equal(sft.streamlines._offsets, offsets_fix)

    @pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
    def test_execution_convert_to_trx(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            in_trk = os.path.join(get_home(), "DSI", "CC_fix.trk")
            exp_data = os.path.join(get_home(), "DSI", "CC_fix_data.npy")
            exp_offsets = os.path.join(get_home(), "DSI", "CC_fix_offsets.npy")
            out_trx_path = os.path.join(tmp_dir, "CC_fix.trx")
            convert_tractogram(in_trk, out_trx_path, None)

            data_fix = np.load(exp_data)
            offsets_fix = np.load(exp_offsets)

            trx = tmm.load(out_trx_path)
            assert_equal(trx.streamlines._data.dtype, np.float32)
            assert_equal(trx.streamlines._offsets.dtype, np.uint32)
            assert_array_equal(trx.streamlines._data, data_fix)
            assert_array_equal(trx.streamlines._offsets, offsets_fix)
            trx.close()

    @pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
    def test_execution_convert_from_trx(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            in_trk = os.path.join(get_home(), "DSI", "CC_fix.trk")
            in_nii = os.path.join(get_home(), "DSI", "CC.nii.gz")
            exp_data = os.path.join(get_home(), "DSI", "CC_fix_data.npy")
            exp_offsets = os.path.join(get_home(), "DSI", "CC_fix_offsets.npy")

            # Sequential conversions
            out_trx_path = os.path.join(tmp_dir, "CC_fix.trx")
            out_trk_path = os.path.join(tmp_dir, "CC_fix.trk")
            out_tck_path = os.path.join(tmp_dir, "CC_fix.tck")
            convert_tractogram(in_trk, out_trx_path, None)
            convert_tractogram(out_trx_path, out_tck_path, None)
            convert_tractogram(out_trx_path, out_trk_path, None)

            data_fix = np.load(exp_data)
            offsets_fix = np.load(exp_offsets)

            sft = load_tractogram(out_trk_path, "same")
            assert_equal(sft.streamlines._data, data_fix)
            assert_equal(sft.streamlines._offsets, offsets_fix)

            sft = load_tractogram(out_tck_path, in_nii)
            assert_equal(sft.streamlines._data, data_fix)
            assert_equal(sft.streamlines._offsets, offsets_fix)

    @pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
    def test_execution_convert_dtype_p16_o64(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            in_trk = os.path.join(get_home(), "DSI", "CC_fix.trk")
            out_convert_path = os.path.join(tmp_dir, "CC_fix_p16_o64.trx")
            convert_tractogram(
                in_trk,
                out_convert_path,
                None,
                pos_dtype="float16",
                offsets_dtype="uint64",
            )

            trx = tmm.load(out_convert_path)
            assert_equal(trx.streamlines._data.dtype, np.float16)
            assert_equal(trx.streamlines._offsets.dtype, np.uint64)
            trx.close()

    @pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
    def test_execution_convert_dtype_p64_o32(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            in_trk = os.path.join(get_home(), "DSI", "CC_fix.trk")
            out_convert_path = os.path.join(tmp_dir, "CC_fix_p16_o64.trx")
            convert_tractogram(
                in_trk,
                out_convert_path,
                None,
                pos_dtype="float64",
                offsets_dtype="uint32",
            )

            trx = tmm.load(out_convert_path)
            assert_equal(trx.streamlines._data.dtype, np.float64)
            assert_equal(trx.streamlines._offsets.dtype, np.uint32)
            trx.close()

    def test_execution_generate_trx_from_scratch(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            reference_fa = os.path.join(get_home(), "trx_from_scratch", "fa.nii.gz")
            raw_arr_dir = os.path.join(get_home(), "trx_from_scratch", "test_npy")
            expected_trx = os.path.join(get_home(), "trx_from_scratch", "expected.trx")

            dpv = [
                (os.path.join(raw_arr_dir, "dpv_cx.npy"), "uint8"),
                (os.path.join(raw_arr_dir, "dpv_cy.npy"), "uint8"),
                (os.path.join(raw_arr_dir, "dpv_cz.npy"), "uint8"),
            ]
            dps = [
                (os.path.join(raw_arr_dir, "dps_algo.npy"), "uint8"),
                (os.path.join(raw_arr_dir, "dps_cw.npy"), "float64"),
            ]
            dpg = [
                (
                    "g_AF_L",
                    os.path.join(raw_arr_dir, "dpg_AF_L_mean_fa.npy"),
                    "float32",
                ),
                (
                    "g_AF_R",
                    os.path.join(raw_arr_dir, "dpg_AF_R_mean_fa.npy"),
                    "float32",
                ),
                ("g_AF_L", os.path.join(raw_arr_dir, "dpg_AF_L_volume.npy"), "float32"),
            ]
            groups = [
                (os.path.join(raw_arr_dir, "g_AF_L.npy"), "int32"),
                (os.path.join(raw_arr_dir, "g_AF_R.npy"), "int32"),
                (os.path.join(raw_arr_dir, "g_CST_L.npy"), "int32"),
            ]

            out_gen_path = os.path.join(tmp_dir, "generated.trx")
            generate_trx_from_scratch(
                reference_fa,
                out_gen_path,
                positions=os.path.join(raw_arr_dir, "positions.npy"),
                offsets=os.path.join(raw_arr_dir, "offsets.npy"),
                positions_dtype="float16",
                offsets_dtype="uint64",
                space_str="rasmm",
                origin_str="nifti",
                verify_invalid=False,
                dpv=dpv,
                dps=dps,
                groups=groups,
                dpg=dpg,
            )
            exp_trx = tmm.load(expected_trx)
            gen_trx = tmm.load(out_gen_path)

            assert DeepDiff(exp_trx.get_dtype_dict(), gen_trx.get_dtype_dict()) == {}

            assert_allclose(
                exp_trx.streamlines._data, gen_trx.streamlines._data, atol=0.1, rtol=0.1
            )
            assert_equal(exp_trx.streamlines._offsets, gen_trx.streamlines._offsets)

            for key in exp_trx.data_per_vertex.keys():
                assert_equal(
                    exp_trx.data_per_vertex[key]._data,
                    gen_trx.data_per_vertex[key]._data,
                )
                assert_equal(
                    exp_trx.data_per_vertex[key]._offsets,
                    gen_trx.data_per_vertex[key]._offsets,
                )
            for key in exp_trx.data_per_streamline.keys():
                assert_equal(
                    exp_trx.data_per_streamline[key], gen_trx.data_per_streamline[key]
                )
            for key in exp_trx.groups.keys():
                assert_equal(exp_trx.groups[key], gen_trx.groups[key])

            for group in exp_trx.groups.keys():
                if group in exp_trx.data_per_group:
                    for key in exp_trx.data_per_group[group].keys():
                        assert_equal(
                            exp_trx.data_per_group[group][key],
                            gen_trx.data_per_group[group][key],
                        )
            exp_trx.close()
            gen_trx.close()

    @pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
    def test_execution_concatenate_validate_trx(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            trx1 = tmm.load(os.path.join(get_home(), "gold_standard", "gs.trx"))
            trx2 = tmm.load(os.path.join(get_home(), "gold_standard", "gs.trx"))
            trx = tmm.concatenate([trx1, trx2], preallocation=False)

            # Right size
            assert_equal(len(trx.streamlines), 2 * len(trx1.streamlines))

            # Right data
            end_idx = trx1.header["NB_VERTICES"]
            assert_allclose(trx.streamlines._data[:end_idx], trx1.streamlines._data)
            assert_allclose(trx.streamlines._data[end_idx:], trx2.streamlines._data)

            # Right data_per_*
            for key in trx.data_per_vertex.keys():
                assert_equal(
                    trx.data_per_vertex[key]._data[:end_idx],
                    trx1.data_per_vertex[key]._data,
                )
                assert_equal(
                    trx.data_per_vertex[key]._data[end_idx:],
                    trx2.data_per_vertex[key]._data,
                )

            end_idx = trx1.header["NB_STREAMLINES"]
            for key in trx.data_per_streamline.keys():
                assert_equal(
                    trx.data_per_streamline[key][:end_idx],
                    trx1.data_per_streamline[key],
                )
                assert_equal(
                    trx.data_per_streamline[key][end_idx:],
                    trx2.data_per_streamline[key],
                )

            # Validate
            out_concat_path = os.path.join(tmp_dir, "concat.trx")
            out_valid_path = os.path.join(tmp_dir, "valid.trx")
            tmm.save(trx, out_concat_path)
            validate_tractogram(
                out_concat_path,
                None,
                out_valid_path,
                remove_identical_streamlines=True,
                precision=0,
            )
            trx_val = tmm.load(out_valid_path)

            # Right dtype and size
            assert DeepDiff(trx.get_dtype_dict(), trx_val.get_dtype_dict()) == {}
            assert_equal(len(trx1.streamlines), len(trx_val.streamlines))

            trx.close()
            trx1.close()
            trx2.close()
            trx_val.close()

    @pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
    def test_execution_manipulate_trx_datatype(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            expected_trx = os.path.join(get_home(), "trx_from_scratch", "expected.trx")
            trx = tmm.load(expected_trx)

            expected_dtype = {
                "positions": np.dtype("float16"),
                "offsets": np.dtype("uint64"),
                "dpv": {
                    "dpv_cx": np.dtype("uint8"),
                    "dpv_cy": np.dtype("uint8"),
                    "dpv_cz": np.dtype("uint8"),
                },
                "dps": {"dps_algo": np.dtype("uint8"), "dps_cw": np.dtype("float64")},
                "dpg": {
                    "g_AF_L": {
                        "dpg_AF_L_mean_fa": np.dtype("float32"),
                        "dpg_AF_L_volume": np.dtype("float32"),
                    },
                    "g_AF_R": {"dpg_AF_R_mean_fa": np.dtype("float32")},
                },
                "groups": {"g_AF_L": np.dtype("int32"), "g_AF_R": np.dtype("int32")},
            }

            assert (
                DeepDiff(
                    trx.get_dtype_dict(),
                    _normalize_dtype_dict(expected_dtype),
                )
                == {}
            )
            trx.close()

            generated_dtype = {
                "positions": np.dtype("float32"),
                "offsets": np.dtype("uint32"),
                "dpv": {
                    "dpv_cx": np.dtype("uint16"),
                    "dpv_cy": np.dtype("uint16"),
                    "dpv_cz": np.dtype("uint16"),
                },
                "dps": {"dps_algo": np.dtype("uint8"), "dps_cw": np.dtype("float32")},
                "dpg": {
                    "g_AF_L": {
                        "dpg_AF_L_mean_fa": np.dtype("float64"),
                        "dpg_AF_L_volume": np.dtype("float32"),
                    },
                    "g_AF_R": {"dpg_AF_R_mean_fa": np.dtype("float64")},
                },
                "groups": {"g_AF_L": np.dtype("uint16"), "g_AF_R": np.dtype("uint16")},
            }

            out_gen_path = os.path.join(tmp_dir, "generated.trx")
            manipulate_trx_datatype(expected_trx, out_gen_path, generated_dtype)
            trx = tmm.load(out_gen_path)
            assert (
                DeepDiff(
                    trx.get_dtype_dict(),
                    _normalize_dtype_dict(generated_dtype),
                )
                == {}
            )
            trx.close()



# -*- coding: utf-8 -*-
"""
TRX Command Line Interface.

This module provides a unified CLI for all TRX file format operations using Typer.
"""

from pathlib import Path
from typing import Annotated, List, Optional

import numpy as np
import typer

from trx.io import load, save
from trx.trx_file_memmap import TrxFile, concatenate, load as load_trx
from trx.workflows import (
    convert_dsi_studio,
    convert_tractogram,
    generate_trx_from_scratch,
    manipulate_trx_datatype,
    tractogram_simple_compare,
    tractogram_visualize_overlap,
    validate_tractogram,
    verify_header_compatibility,
)


def _debug_callback(value: bool) -> None:
    """Print environment and dependency diagnostics, then exit.

    Parameters
    ----------
    value : bool
        Whether the ``--debug`` flag was passed.
    """
    if not value:
        return

    import importlib.metadata
    import importlib.util
    import sys

    from trx import __version__

    typer.echo("Environment diagnostics:")
    typer.echo(f"  Python executable : {sys.executable}")
    typer.echo(f"  sys.prefix        : {sys.prefix}")
    typer.echo(f"  trx-python version: {__version__}")

    trx_spec = importlib.util.find_spec("trx")
    trx_location = (
        trx_spec.submodule_search_locations[0]
        if trx_spec and trx_spec.submodule_search_locations
        else "unknown"
    )
    typer.echo(f"  trx package       : {trx_location}")

    # Read required dependencies from package metadata
    required_deps = []
    try:
        import re

        for req in importlib.metadata.requires("trx-python") or []:
            # Skip optional / extra deps (they contain "; extra ==")
            if "extra ==" in req:
                continue
            # Extract the package name (strip version specifiers like >=, <=, ~=)
            dep_name = re.split(r"[>=<!~;\s]", req)[0]
            required_deps.append(dep_name)
    except importlib.metadata.PackageNotFoundError:
        required_deps = ["deepdiff", "nibabel", "numpy", "typer"]

    optional_deps = ["dipy", "fury", "vtk"]

    typer.echo("\nRequired dependencies:")
    for dep in required_deps:
        spec = importlib.util.find_spec(dep)
        if spec is None:
            typer.echo(f"  {dep:12s} NOT FOUND")
        else:
            typer.echo(f"  {dep:12s} found")

    typer.echo("\nOptional dependencies:")
    for dep in optional_deps:
        spec = importlib.util.find_spec(dep)
        if spec is None:
            typer.echo(f"  {dep:12s} not found")
        else:
            typer.echo(f"  {dep:12s} found")

    raise typer.Exit()


def _main_callback(
    _debug: Annotated[
        bool,
        typer.Option(
            "--debug",
            help="Print environment and dependency diagnostics.",
            callback=_debug_callback,
            is_eager=True,
        ),
    ] = False,
) -> None:
    """TRX File Format Tools - CLI for brain tractography data manipulation.

    Parameters
    ----------
    _debug : bool, optional
        If True, print environment and dependency diagnostics and exit.
    """


app = typer.Typer(
    name="trx",
    help="TRX File Format Tools - CLI for brain tractography data manipulation.",
    add_completion=False,
    rich_markup_mode="rich",
    callback=_main_callback,
)


def _check_overwrite(filepath: Path, overwrite: bool) -> None:
    """Check if file exists and raise error if overwrite is not enabled.

    Parameters
    ----------
    filepath : Path
        Path to the output file.
    overwrite : bool
        If True, allow overwriting existing files.

    Raises
    ------
    typer.Exit
        If file exists and overwrite is False.
    """
    if filepath.is_file() and not overwrite:
        typer.echo(
            typer.style(
                f"Error: {filepath} already exists. Use --force to overwrite.",
                fg=typer.colors.RED,
            ),
            err=True,
        )
        raise typer.Exit(code=1)


@app.command("concatenate")
def concatenate_tractograms(
    in_tractograms: Annotated[
        List[Path],
        typer.Argument(
            help="Input tractogram files. Format: trk, tck, vtk, fib, dpy, trx.",
        ),
    ],
    out_tractogram: Annotated[
        Path,
        typer.Argument(help="Output filename for the concatenated tractogram."),
    ],
    delete_dpv: Annotated[
        bool,
        typer.Option(
            "--delete-dpv",
            help="Delete data_per_vertex if not all inputs have the same metadata.",
        ),
    ] = False,
    delete_dps: Annotated[
        bool,
        typer.Option(
            "--delete-dps",
            help="Delete data_per_streamline if not all inputs have the same metadata.",
        ),
    ] = False,
    delete_groups: Annotated[
        bool,
        typer.Option(
            "--delete-groups",
            help="Delete groups if not all inputs have the same metadata.",
        ),
    ] = False,
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference anatomy for tck/vtk/fib/dpy files (.nii or .nii.gz).",
        ),
    ] = None,
    force: Annotated[
        bool,
        typer.Option("--force", "-f", help="Force overwriting of output files."),
    ] = False,
) -> None:
    """Concatenate multiple tractograms into one.

    If the data_per_point or data_per_streamline is not the same for all
    tractograms, the data must be deleted first using the appropriate flags.

    Parameters
    ----------
    in_tractograms : list of Path
        Input tractogram files (.trk, .tck, .vtk, .fib, .dpy, .trx).
    out_tractogram : Path
        Output filename for the concatenated tractogram.
    delete_dpv : bool, optional
        Delete ``data_per_vertex`` if metadata differ across inputs.
    delete_dps : bool, optional
        Delete ``data_per_streamline`` if metadata differ across inputs.
    delete_groups : bool, optional
        Delete groups when metadata differ across inputs.
    reference : Path or None, optional
        Reference anatomy for tck/vtk/fib/dpy inputs.
    force : bool, optional
        Overwrite output if it already exists.

    Returns
    -------
    None
        Writes the concatenated tractogram to ``out_tractogram``.
    """
    _check_overwrite(out_tractogram, force)

    ref = str(reference) if reference else None

    trx_list = []
    has_group = False
    for filename in in_tractograms:
        tractogram_obj = load(str(filename), ref)

        if not isinstance(tractogram_obj, TrxFile):
            tractogram_obj = TrxFile.from_sft(tractogram_obj)
        elif len(tractogram_obj.groups):
            has_group = True
        trx_list.append(tractogram_obj)

    trx = concatenate(
        trx_list,
        delete_dpv=delete_dpv,
        delete_dps=delete_dps,
        delete_groups=delete_groups or not has_group,
        check_space_attributes=True,
        preallocation=False,
    )
    save(trx, str(out_tractogram))

    typer.echo(
        typer.style(
            f"Successfully concatenated {len(in_tractograms)} tractograms "
            f"to {out_tractogram}",
            fg=typer.colors.GREEN,
        )
    )


@app.command("convert")
def convert(
    in_tractogram: Annotated[
        Path,
        typer.Argument(help="Input tractogram. Format: trk, tck, vtk, fib, dpy, trx."),
    ],
    out_tractogram: Annotated[
        Path,
        typer.Argument(help="Output tractogram. Format: trk, tck, vtk, fib, dpy, trx."),
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference anatomy for tck/vtk/fib/dpy files (.nii or .nii.gz).",
        ),
    ] = None,
    positions_dtype: Annotated[
        str,
        typer.Option(
            "--positions-dtype",
            help="Datatype for positions in TRX output.",
        ),
    ] = "float32",
    offsets_dtype: Annotated[
        str,
        typer.Option(
            "--offsets-dtype",
            help="Datatype for offsets in TRX output.",
        ),
    ] = "uint64",
    force: Annotated[
        bool,
        typer.Option("--force", "-f", help="Force overwriting of output files."),
    ] = False,
) -> None:
    """Convert tractograms between formats.

    Supports conversion of .tck, .trk, .fib, .vtk, .trx and .dpy files.
    TCK files always need a reference NIFTI file for conversion.

    Parameters
    ----------
    in_tractogram : Path
        Input tractogram file.
    out_tractogram : Path
        Output tractogram path.
    reference : Path or None, optional
        Reference anatomy required for some input formats.
    positions_dtype : str, optional
        Datatype for positions in TRX output.
    offsets_dtype : str, optional
        Datatype for offsets in TRX output.
    force : bool, optional
        Overwrite output if it already exists.

    Returns
    -------
    None
        Writes the converted tractogram to disk.
    """
    _check_overwrite(out_tractogram, force)

    ref = str(reference) if reference else None
    convert_tractogram(
        str(in_tractogram),
        str(out_tractogram),
        ref,
        pos_dtype=positions_dtype,
        offsets_dtype=offsets_dtype,
    )

    typer.echo(
        typer.style(
            f"Successfully converted {in_tractogram} to {out_tractogram}",
            fg=typer.colors.GREEN,
        )
    )


@app.command("convert-dsi")
def convert_dsi(
    in_dsi_tractogram: Annotated[
        Path,
        typer.Argument(help="Input tractogram from DSI Studio (.trk)."),
    ],
    in_dsi_fa: Annotated[
        Path,
        typer.Argument(help="Input FA from DSI Studio (.nii.gz)."),
    ],
    out_tractogram: Annotated[
        Path,
        typer.Argument(help="Output tractogram file."),
    ],
    remove_invalid: Annotated[
        bool,
        typer.Option(
            "--remove-invalid",
            help="Remove streamlines landing out of the bounding box.",
        ),
    ] = False,
    keep_invalid: Annotated[
        bool,
        typer.Option(
            "--keep-invalid",
            help="Keep streamlines landing out of the bounding box.",
        ),
    ] = False,
    force: Annotated[
        bool,
        typer.Option("--force", "-f", help="Force overwriting of output files."),
    ] = False,
) -> None:
    """Convert a DSI-Studio TRK file to TRX or TRK and fix space metadata.

    Parameters
    ----------
    in_dsi_tractogram : Path
        Input DSI-Studio tractogram (.trk or .trk.gz).
    in_dsi_fa : Path
        FA volume used as reference (.nii.gz).
    out_tractogram : Path
        Output tractogram path (.trx or .trk).
    remove_invalid : bool, optional
        Remove streamlines outside the bounding box. Defaults to False.
    keep_invalid : bool, optional
        Keep streamlines outside the bounding box. Defaults to False.
    force : bool, optional
        Overwrite output if it already exists.

    Returns
    -------
    None
        Writes the converted tractogram to disk.
    """
    _check_overwrite(out_tractogram, force)

    if remove_invalid and keep_invalid:
        typer.echo(
            typer.style(
                "Error: Cannot use both --remove-invalid and --keep-invalid.",
                fg=typer.colors.RED,
            ),
            err=True,
        )
        raise typer.Exit(code=1)

    convert_dsi_studio(
        str(in_dsi_tractogram),
        str(in_dsi_fa),
        str(out_tractogram),
        remove_invalid=remove_invalid,
        keep_invalid=keep_invalid,
    )

    typer.echo(
        typer.style(
            f"Successfully converted DSI-Studio tractogram to {out_tractogram}",
            fg=typer.colors.GREEN,
        )
    )


@app.command("generate")
def generate(
    reference: Annotated[
        Path,
        typer.Argument(help="Reference anatomy (.nii or .nii.gz)."),
    ],
    out_tractogram: Annotated[
        Path,
        typer.Argument(help="Output tractogram. Format: trk, tck, vtk, fib, dpy, trx."),
    ],
    positions: Annotated[
        Optional[Path],
        typer.Option(
            "--positions",
            help="Binary file with streamline coordinates (Nx3 .npy).",
        ),
    ] = None,
    offsets: Annotated[
        Optional[Path],
        typer.Option(
            "--offsets",
            help="Binary file with streamline offsets (.npy).",
        ),
    ] = None,
    positions_csv: Annotated[
        Optional[Path],
        typer.Option(
            "--positions-csv",
            help="CSV file with streamline coordinates (x1,y1,z1,x2,y2,z2,...).",
        ),
    ] = None,
    space: Annotated[
        str,
        typer.Option(
            "--space",
            help="Coordinate space. Non-default requires Dipy.",
        ),
    ] = "RASMM",
    origin: Annotated[
        str,
        typer.Option(
            "--origin",
            help="Coordinate origin. Non-default requires Dipy.",
        ),
    ] = "NIFTI",
    positions_dtype: Annotated[
        str,
        typer.Option("--positions-dtype", help="Datatype for positions."),
    ] = "float32",
    offsets_dtype: Annotated[
        str,
        typer.Option("--offsets-dtype", help="Datatype for offsets."),
    ] = "uint64",
    dpv: Annotated[
        Optional[List[str]],
        typer.Option(
            "--dpv",
            help="Data per vertex: FILE,DTYPE (e.g., color.npy,uint8).",
        ),
    ] = None,
    dps: Annotated[
        Optional[List[str]],
        typer.Option(
            "--dps",
            help="Data per streamline: FILE,DTYPE (e.g., algo.npy,uint8).",
        ),
    ] = None,
    groups: Annotated[
        Optional[List[str]],
        typer.Option(
            "--groups",
            help="Groups: FILE,DTYPE (e.g., AF_L.npy,int32).",
        ),
    ] = None,
    dpg: Annotated[
        Optional[List[str]],
        typer.Option(
            "--dpg",
            help="Data per group: GROUP,FILE,DTYPE (e.g., AF_L,mean_fa.npy,float32).",
        ),
    ] = None,
    verify_invalid: Annotated[
        bool,
        typer.Option(
            "--verify-invalid",
            help="Verify positions are valid (within bounding box). Requires Dipy.",
        ),
    ] = False,
    force: Annotated[
        bool,
        typer.Option("--force", "-f", help="Force overwriting of output files."),
    ] = False,
) -> None:
    """Generate a TRX file from raw data files.

    Create a TRX file from CSV, TXT, or NPY files by specifying positions,
    offsets, data_per_vertex, data_per_streamlines, groups, and data_per_group.

    Parameters
    ----------
    reference : Path
        Reference anatomy (.nii or .nii.gz).
    out_tractogram : Path
        Output tractogram (.trk, .tck, .vtk, .fib, .dpy, .trx).
    positions : Path or None, optional
        Binary file with streamline coordinates (Nx3 .npy).
    offsets : Path or None, optional
        Binary file with streamline offsets (.npy).
    positions_csv : Path or None, optional
        CSV file with flattened streamline coordinates.
    space : str, optional
        Coordinate space. Non-default requires Dipy.
    origin : str, optional
        Coordinate origin. Non-default requires Dipy.
    positions_dtype : str, optional
        Datatype for positions.
    offsets_dtype : str, optional
        Datatype for offsets.
    dpv : list of str or None, optional
        Data per vertex entries as FILE,DTYPE pairs.
    dps : list of str or None, optional
        Data per streamline entries as FILE,DTYPE pairs.
    groups : list of str or None, optional
        Group entries as FILE,DTYPE pairs.
    dpg : list of str or None, optional
        Data per group entries as GROUP,FILE,DTYPE triplets.
    verify_invalid : bool, optional
        Verify positions are inside bounding box (requires Dipy).
    force : bool, optional
        Overwrite output if it already exists.

    Returns
    -------
    None
        Writes the generated tractogram to disk.
    """
    _check_overwrite(out_tractogram, force)

    # Validate input combinations
    if not positions and not positions_csv:
        typer.echo(
            typer.style(
                "Error: At least one positions option must be provided "
                "(--positions or --positions-csv).",
                fg=typer.colors.RED,
            ),
            err=True,
        )
        raise typer.Exit(code=1)
    if positions_csv and positions:
        typer.echo(
            typer.style(
                "Error: Cannot use both --positions and --positions-csv.",
                fg=typer.colors.RED,
            ),
            err=True,
        )
        raise typer.Exit(code=1)
    if positions and offsets is None:
        typer.echo(
            typer.style(
                "Error: --offsets must be provided if --positions is used.",
                fg=typer.colors.RED,
            ),
            err=True,
        )
        raise typer.Exit(code=1)
    if offsets and positions is None:
        typer.echo(
            typer.style(
                "Error: --positions must be provided if --offsets is used.",
                fg=typer.colors.RED,
            ),
            err=True,
        )
        raise typer.Exit(code=1)

    # Parse comma-separated arguments to tuples
    dpv_list = None
    if dpv:
        dpv_list = [tuple(item.split(",")) for item in dpv]

    dps_list = None
    if dps:
        dps_list = [tuple(item.split(",")) for item in dps]

    groups_list = None
    if groups:
        groups_list = [tuple(item.split(",")) for item in groups]

    dpg_list = None
    if dpg:
        dpg_list = [tuple(item.split(",")) for item in dpg]

    generate_trx_from_scratch(
        str(reference),
        str(out_tractogram),
        positions_csv=str(positions_csv) if positions_csv else None,
        positions=str(positions) if positions else None,
        offsets=str(offsets) if offsets else None,
        positions_dtype=positions_dtype,
        offsets_dtype=offsets_dtype,
        space_str=space,
        origin_str=origin,
        verify_invalid=verify_invalid,
        dpv=dpv_list,
        dps=dps_list,
        groups=groups_list,
        dpg=dpg_list,
    )

    typer.echo(
        typer.style(
            f"Successfully generated {out_tractogram}",
            fg=typer.colors.GREEN,
        )
    )


@app.command("manipulate-dtype")
def manipulate_dtype(
    in_tractogram: Annotated[
        Path,
        typer.Argument(help="Input TRX file."),
    ],
    out_tractogram: Annotated[
        Path,
        typer.Argument(help="Output tractogram file."),
    ],
    positions_dtype: Annotated[
        Optional[str],
        typer.Option(
            "--positions-dtype",
            help="Datatype for positions (float16, float32, float64).",
        ),
    ] = None,
    offsets_dtype: Annotated[
        Optional[str],
        typer.Option(
            "--offsets-dtype",
            help="Datatype for offsets (uint32, uint64).",
        ),
    ] = None,
    dpv: Annotated[
        Optional[List[str]],
        typer.Option(
            "--dpv",
            help="Data per vertex dtype: NAME,DTYPE (e.g., color_x,uint8).",
        ),
    ] = None,
    dps: Annotated[
        Optional[List[str]],
        typer.Option(
            "--dps",
            help="Data per streamline dtype: NAME,DTYPE (e.g., algo,uint8).",
        ),
    ] = None,
    groups: Annotated[
        Optional[List[str]],
        typer.Option(
            "--groups",
            help="Groups dtype: NAME,DTYPE (e.g., CC,uint64).",
        ),
    ] = None,
    dpg: Annotated[
        Optional[List[str]],
        typer.Option(
            "--dpg",
            help="Data per group dtype: GROUP,NAME,DTYPE (e.g., CC,mean_fa,float64).",
        ),
    ] = None,
    force: Annotated[
        bool,
        typer.Option("--force", "-f", help="Force overwriting of output files."),
    ] = False,
) -> None:  # noqa: C901
    """Manipulate TRX file internal array data types.

    Change the data types of positions, offsets, data_per_vertex,
    data_per_streamline, groups, and data_per_group arrays.

    Parameters
    ----------
    in_tractogram : Path
        Input TRX file.
    out_tractogram : Path
        Output TRX file.
    positions_dtype : str or None, optional
        Target dtype for positions (float16, float32, float64).
    offsets_dtype : str or None, optional
        Target dtype for offsets (uint32, uint64).
    dpv : list of str or None, optional
        Data per vertex dtype overrides as NAME,DTYPE pairs.
    dps : list of str or None, optional
        Data per streamline dtype overrides as NAME,DTYPE pairs.
    groups : list of str or None, optional
        Group dtype overrides as NAME,DTYPE pairs.
    dpg : list of str or None, optional
        Data per group dtype overrides as GROUP,NAME,DTYPE triplets.
    force : bool, optional
        Overwrite output if it already exists.

    Returns
    -------
    None
        Writes the dtype-converted TRX file.
    """
    _check_overwrite(out_tractogram, force)

    dtype_dict = {}
    if positions_dtype:
        dtype_dict["positions"] = np.dtype(positions_dtype)
    if offsets_dtype:
        dtype_dict["offsets"] = np.dtype(offsets_dtype)
    if dpv:
        dtype_dict["dpv"] = {}
        for item in dpv:
            name, dtype = item.split(",")
            dtype_dict["dpv"][name] = np.dtype(dtype)
    if dps:
        dtype_dict["dps"] = {}
        for item in dps:
            name, dtype = item.split(",")
            dtype_dict["dps"][name] = np.dtype(dtype)
    if groups:
        dtype_dict["groups"] = {}
        for item in groups:
            name, dtype = item.split(",")
            dtype_dict["groups"][name] = np.dtype(dtype)
    if dpg:
        dtype_dict["dpg"] = {}
        for item in dpg:
            parts = item.split(",")
            group, name, dtype = parts[0], parts[1], parts[2]
            if group not in dtype_dict["dpg"]:
                dtype_dict["dpg"][group] = {}
            dtype_dict["dpg"][group][name] = np.dtype(dtype)

    manipulate_trx_datatype(str(in_tractogram), str(out_tractogram), dtype_dict)

    typer.echo(
        typer.style(
            f"Successfully manipulated datatypes and saved to {out_tractogram}",
            fg=typer.colors.GREEN,
        )
    )


@app.command("compare")
def compare(
    in_tractogram1: Annotated[
        Path,
        typer.Argument(help="First tractogram file."),
    ],
    in_tractogram2: Annotated[
        Path,
        typer.Argument(help="Second tractogram file."),
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference anatomy for tck/vtk/fib/dpy files (.nii or .nii.gz).",
        ),
    ] = None,
) -> None:
    """Compare two tractograms and report basic differences.

    Parameters
    ----------
    in_tractogram1 : Path
        First tractogram file.
    in_tractogram2 : Path
        Second tractogram file.
    reference : Path or None, optional
        Reference anatomy for formats requiring it.

    Returns
    -------
    None
        Prints comparison summary to stdout.
    """
    ref = str(reference) if reference else None
    tractogram_simple_compare([str(in_tractogram1), str(in_tractogram2)], ref)


@app.command("validate")
def validate(
    in_tractogram: Annotated[
        Path,
        typer.Argument(help="Input tractogram. Format: trk, tck, vtk, fib, dpy, trx."),
    ],
    out_tractogram: Annotated[
        Optional[Path],
        typer.Option(
            "--out",
            "-o",
            help="Output tractogram after removing invalid streamlines.",
        ),
    ] = None,
    remove_identical: Annotated[
        bool,
        typer.Option(
            "--remove-identical",
            help="Remove identical streamlines from the set.",
        ),
    ] = False,
    precision: Annotated[
        int,
        typer.Option(
            "--precision",
            "-p",
            help="Number of decimals when hashing streamline points.",
        ),
    ] = 1,
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference anatomy for tck/vtk/fib/dpy files (.nii or .nii.gz).",
        ),
    ] = None,
    force: Annotated[
        bool,
        typer.Option("--force", "-f", help="Force overwriting of output files."),
    ] = False,
) -> None:
    """Validate a tractogram and optionally clean invalid/duplicate streamlines.

    Parameters
    ----------
    in_tractogram : Path
        Input tractogram (.trk, .tck, .vtk, .fib, .dpy, .trx).
    out_tractogram : Path or None, optional
        Optional output tractogram with invalid streamlines removed.
    remove_identical : bool, optional
        Remove duplicate streamlines based on hashing precision.
    precision : int, optional
        Number of decimals when hashing streamline points.
    reference : Path or None, optional
        Reference anatomy for formats requiring it.
    force : bool, optional
        Overwrite output if it already exists.

    Returns
    -------
    None
        Prints validation summary and optionally writes cleaned output.
    """
    if out_tractogram:
        _check_overwrite(out_tractogram, force)

    ref = str(reference) if reference else None
    out = str(out_tractogram) if out_tractogram else None

    validate_tractogram(
        str(in_tractogram),
        reference=ref,
        out_tractogram=out,
        remove_identical_streamlines=remove_identical,
        precision=precision,
    )

    if out_tractogram:
        typer.echo(
            typer.style(
                f"Validation complete. Output saved to {out_tractogram}",
                fg=typer.colors.GREEN,
            )
        )
    else:
        typer.echo(
            typer.style(
                "Validation complete.",
                fg=typer.colors.GREEN,
            )
        )


@app.command("verify-header")
def verify_header(
    in_files: Annotated[
        List[Path],
        typer.Argument(help="Files to compare (trk, trx, and nii)."),
    ],
) -> None:
    """Compare spatial attributes of input files.

    Parameters
    ----------
    in_files : list of Path
        Files to compare (.trk, .trx, .nii, .nii.gz).

    Returns
    -------
    None
        Prints compatibility results to stdout.
    """
    verify_header_compatibility([str(f) for f in in_files])


@app.command("visualize")
def visualize(
    in_tractogram: Annotated[
        Path,
        typer.Argument(help="Input tractogram. Format: trk, tck, vtk, fib, dpy, trx."),
    ],
    reference: Annotated[
        Path,
        typer.Argument(help="Reference anatomy (.nii or .nii.gz)."),
    ],
    remove_invalid: Annotated[
        bool,
        typer.Option(
            "--remove-invalid",
            help="Remove invalid streamlines to avoid density_map crash.",
        ),
    ] = False,
) -> None:
    """Display tractogram and density map with bounding box.

    Parameters
    ----------
    in_tractogram : Path
        Input tractogram (.trk, .tck, .vtk, .fib, .dpy, .trx).
    reference : Path
        Reference anatomy (.nii or .nii.gz).
    remove_invalid : bool, optional
        Remove invalid streamlines to avoid density map crashes.

    Returns
    -------
    None
        Opens visualization windows when fury is available.
    """
    tractogram_visualize_overlap(
        str(in_tractogram),
        str(reference),
        remove_invalid,
    )


def _format_size(size_bytes: int) -> str:
    """Format byte size to human readable string.

    Parameters
    ----------
    size_bytes : int
        Size in bytes.

    Returns
    -------
    str
        Human readable size string (e.g., "1.5 MB").
    """
    for unit in ["B", "KB", "MB", "GB"]:
        if size_bytes < 1024:
            return f"{size_bytes:.1f} {unit}" if unit != "B" else f"{size_bytes} {unit}"
        size_bytes /= 1024
    return f"{size_bytes:.1f} TB"


@app.command("info")
def info(
    in_tractogram: Annotated[
        Path,
        typer.Argument(help="Input TRX file."),
    ],
) -> None:
    """Display detailed information about a TRX file.

    Shows file size, compression status, header metadata (affine, dimensions,
    voxel sizes), streamline/vertex counts, data keys (dpv, dps, dpg), groups,
    and archive contents listing similar to ``unzip -l``.

    Parameters
    ----------
    in_tractogram : Path
        Input TRX file (.trx extension required).

    Returns
    -------
    None
        Prints TRX file information to stdout.

    Examples
    --------
    $ trx info tractogram.trx
    $ trx_info tractogram.trx
    """
    import zipfile

    if not in_tractogram.exists():
        typer.echo(
            typer.style(f"Error: {in_tractogram} does not exist.", fg=typer.colors.RED),
            err=True,
        )
        raise typer.Exit(code=1)

    if in_tractogram.suffix.lower() != ".trx":
        typer.echo(
            typer.style(
                f"Error: {in_tractogram.name} is not a TRX file. "
                "Only .trx files are supported.",
                fg=typer.colors.RED,
            ),
            err=True,
        )
        raise typer.Exit(code=1)

    # Show archive info
    file_size = in_tractogram.stat().st_size
    typer.echo(f"File: {in_tractogram}")
    typer.echo(f"Size: {_format_size(file_size)}")

    with zipfile.ZipFile(str(in_tractogram), "r") as zf:
        total_uncompressed = sum(info.file_size for info in zf.infolist())
        is_compressed = any(info.compress_type != 0 for info in zf.infolist())
        typer.echo(f"Entries: {len(zf.infolist())}")
        typer.echo(f"Compressed: {'Yes' if is_compressed else 'No'}")
        typer.echo(f"Uncompressed size: {_format_size(total_uncompressed)}")

    typer.echo("")

    # Show TRX content info
    trx = load_trx(str(in_tractogram))
    typer.echo(trx)

    # Show file listing (unzip -l style)
    typer.echo("\nArchive contents:")
    typer.echo("  Length      Date    Time    Name")
    typer.echo("---------  ---------- -----   ----")
    with zipfile.ZipFile(str(in_tractogram), "r") as zf:
        for zinfo in zf.infolist():
            dt = zinfo.date_time
            date_str = f"{dt[1]:02d}-{dt[2]:02d}-{dt[0]}"
            time_str = f"{dt[3]:02d}:{dt[4]:02d}"
            typer.echo(
                f"{zinfo.file_size:>9}  {date_str} {time_str}   {zinfo.filename}"
            )
        num_files = len(zf.infolist())
        typer.echo("---------                     -------")
        typer.echo(f"{total_uncompressed:>9}                     {num_files} files")

    trx.close()


def main():
    """Entry point for the TRX CLI."""
    app()


# Standalone entry points for backward compatibility
# These create individual Typer apps for each command


def _create_standalone_app(command_func, name: str, help_text: str):
    """Create a standalone Typer app for a single command.

    Parameters
    ----------
    command_func : callable
        The command function to wrap.
    name : str
        Name of the command.
    help_text : str
        Help text for the command.

    Returns
    -------
    callable
        Entry point function.
    """
    standalone = typer.Typer(
        name=name,
        help=help_text,
        add_completion=False,
        rich_markup_mode="rich",
    )
    standalone.command()(command_func)
    return lambda: standalone()


concatenate_tractograms_cmd = _create_standalone_app(
    concatenate_tractograms,
    "trx_concatenate_tractograms",
    "Concatenate multiple tractograms into one.",
)

convert_dsi_cmd = _create_standalone_app(
    convert_dsi,
    "trx_convert_dsi_studio",
    "Fix DSI-Studio TRK files for compatibility.",
)

convert_cmd = _create_standalone_app(
    convert,
    "trx_convert_tractogram",
    "Convert tractograms between formats.",
)

generate_cmd = _create_standalone_app(
    generate,
    "trx_generate_from_scratch",
    "Generate TRX file from raw data files.",
)

manipulate_dtype_cmd = _create_standalone_app(
    manipulate_dtype,
    "trx_manipulate_datatype",
    "Manipulate TRX file internal array data types.",
)

compare_cmd = _create_standalone_app(
    compare,
    "trx_simple_compare",
    "Simple comparison of tractograms by subtracting coordinates.",
)

validate_cmd = _create_standalone_app(
    validate,
    "trx_validate",
    "Validate TRX file and remove invalid streamlines.",
)

verify_header_cmd = _create_standalone_app(
    verify_header,
    "trx_verify_header_compatibility",
    "Compare spatial attributes of input files.",
)

visualize_cmd = _create_standalone_app(
    visualize,
    "trx_visualize_overlap",
    "Display tractogram and density map with bounding box.",
)

info_cmd = _create_standalone_app(
    info,
    "trx_info",
    "Display information about a TRX file.",
)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""Set operations on streamlines with precision-based matching."""

from functools import reduce
import itertools

import numpy as np

MIN_NB_POINTS = 5
KEY_INDEX = np.concatenate((range(5), range(-1, -6, -1)))


def intersection(left, right):
    """Return the intersection of two streamline hash dictionaries.

    Parameters
    ----------
    left : dict
        Hash dictionary returned by :func:`hash_streamlines`.
    right : dict
        Hash dictionary returned by :func:`hash_streamlines`.

    Returns
    -------
    dict
        Dictionary containing only keys present in both inputs.
    """
    return {k: v for k, v in left.items() if k in right}


def difference(left, right):
    """Return the difference of two streamline hash dictionaries.

    Parameters
    ----------
    left : dict
        Hash dictionary returned by :func:`hash_streamlines`.
    right : dict
        Hash dictionary returned by :func:`hash_streamlines`.

    Returns
    -------
    dict
        Dictionary containing keys present in ``left`` but not in ``right``.
    """
    return {k: v for k, v in left.items() if k not in right}


def union(left, right):
    """Return the union of two streamline hash dictionaries.

    Parameters
    ----------
    left : dict
        Hash dictionary returned by :func:`hash_streamlines`.
    right : dict
        Hash dictionary returned by :func:`hash_streamlines`.

    Returns
    -------
    dict
        Dictionary containing all keys from both inputs. Values from ``left``
        overwrite those from ``right`` when keys overlap.
    """
    result = right.copy()
    result.update(left)
    return result


def get_streamline_key(streamline, precision=None):
    """Produce a hash key from a streamline using a few points.

    Parameters
    ----------
    streamline : ndarray
        A single streamline (N, 3).
    precision : int, optional
        The number of decimals to keep when hashing the points of the
        streamlines. Allows a soft comparison of streamlines. If None, no
        rounding is performed.

    Returns
    -------
    bytes
        Hash of the first/last MIN_NB_POINTS points of the streamline.
    """

    # Use just a few data points as hash key. I could use all the data of
    # the streamlines, but then the complexity grows with the number of
    # points.
    if len(streamline) < MIN_NB_POINTS:
        key = streamline.copy()
    else:
        key = streamline[KEY_INDEX].copy()

    if precision is not None:
        key = np.round(key, precision)

    key.flags.writeable = False

    return key.data.tobytes()


def hash_streamlines(streamlines, start_index=0, precision=None):
    """Produce a dict from streamlines.

    Produce a dict from streamlines by using the points as keys and the
    indices of the streamlines as values.

    Parameters
    ----------
    streamlines : list of ndarray
        The list of streamlines used to produce the dict.
    start_index : int, optional
        The index of the first streamline. 0 by default.
    precision : int, optional
        The number of decimals to keep when hashing the points of the
        streamlines. Allows a soft comparison of streamlines. If None, no
        rounding is performed.

    Returns
    -------
    dict
        A dict where the keys are streamline points and the values are
        indices starting at start_index.
    """
    keys = [get_streamline_key(s, precision) for s in streamlines]
    return {k: i for i, k in enumerate(keys, start_index)}


def perform_streamlines_operation(operation, streamlines, precision=0):
    """Perform an operation on a list of list of streamlines.

    Given a list of list of streamlines, this function applies the operation
    to the first two lists of streamlines. The result in then used recursively
    with the third, fourth, etc. lists of streamlines.

    A valid operation is any function that takes two streamlines dict as input
    and produces a new streamlines dict (see hash_streamlines). Union,
    difference, and intersection are valid examples of operations.

    Parameters
    ----------
    operation : callable
        A callable that takes two streamlines dicts as inputs and produces a
        new streamline dict.
    streamlines : list of list of streamlines
        The streamlines used in the operation.
    precision : int, optional
        The number of decimals to keep when hashing the points of the
        streamlines. Allows a soft comparison of streamlines. If None, no
        rounding is performed.

    Returns
    -------
    streamlines : list of `nib.streamline.ArraySequence`
        The streamlines obtained after performing the operation on all the
        input streamlines.
    indices : np.ndarray
        The indices of the streamlines that are used in the output.
    """

    # Hash the streamlines using the desired precision.
    indices = np.cumsum([0] + [len(s) for s in streamlines[:-1]])
    hashes = [hash_streamlines(s, i, precision) for s, i in zip(streamlines, indices)]

    # Perform the operation on the hashes and get the output streamlines.
    to_keep = reduce(operation, hashes)
    all_streamlines = list(itertools.chain(*streamlines))
    indices = np.array(sorted(to_keep.values())).astype(np.uint32)
    streamlines = [all_streamlines[i] for i in indices]

    return streamlines, indices



# -*- coding: utf-8 -*-

from copy import deepcopy
import os
from tempfile import TemporaryDirectory
import zipfile

import numpy as np
from numpy.testing import assert_allclose
import psutil
import pytest

try:
    from dipy.io.streamline import load_tractogram, save_tractogram

    dipy_available = True
except ImportError:
    dipy_available = False

from trx.fetcher import fetch_data, get_home, get_testing_files_dict
from trx.io import load, save
import trx.trx_file_memmap as tmm
from trx.trx_file_memmap import TrxFile

fetch_data(get_testing_files_dict(), keys=["gold_standard.zip"])


@pytest.mark.parametrize("path", [("gs.trk"), ("gs.tck"), ("gs.vtk")])
@pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
def test_seq_ops_sft(path):
    with TemporaryDirectory() as tmp_dir:
        gs_dir = os.path.join(get_home(), "gold_standard")
        path = os.path.join(tmp_dir, path)

        obj = load(os.path.join(gs_dir, "gs.trx"), os.path.join(gs_dir, "gs.nii"))
        sft_1 = obj.to_sft()
        save_tractogram(sft_1, path)
        obj.close()
        save_tractogram(sft_1, os.path.join(tmp_dir, "tmp.trk"))

        _ = load_tractogram(os.path.join(tmp_dir, "tmp.trk"), "same")


def test_seq_ops_trx():
    with TemporaryDirectory() as tmp_dir:
        gs_dir = os.path.join(get_home(), "gold_standard")
        path = os.path.join(gs_dir, "gs.trx")

        trx_1 = tmm.load(path)
        tmm.save(trx_1, os.path.join(tmp_dir, "tmp.trx"))
        trx_1.close()
        trx_2 = tmm.load(os.path.join(tmp_dir, "tmp.trx"))
        trx_2.close()


@pytest.mark.parametrize("path", [("gs.trx"), ("gs.trk"), ("gs.tck"), ("gs.vtk")])
@pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
def test_load_vox(path):
    gs_dir = os.path.join(get_home(), "gold_standard")
    path = os.path.join(gs_dir, path)
    coord = np.loadtxt(os.path.join(get_home(), "gold_standard", "gs_vox_space.txt"))
    obj = load(path, os.path.join(gs_dir, "gs.nii"))

    sft = obj.to_sft() if isinstance(obj, TrxFile) else obj
    sft.to_vox()

    assert_allclose(sft.streamlines._data, coord, rtol=1e-04, atol=1e-06)
    if isinstance(obj, TrxFile):
        obj.close()


@pytest.mark.parametrize("path", [("gs.trx"), ("gs.trk"), ("gs.tck"), ("gs.vtk")])
@pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
def test_load_voxmm(path):
    gs_dir = os.path.join(get_home(), "gold_standard")
    path = os.path.join(gs_dir, path)
    coord = np.loadtxt(os.path.join(get_home(), "gold_standard", "gs_voxmm_space.txt"))
    obj = load(path, os.path.join(gs_dir, "gs.nii"))

    sft = obj.to_sft() if isinstance(obj, TrxFile) else obj
    sft.to_voxmm()

    assert_allclose(sft.streamlines._data, coord, rtol=1e-04, atol=1e-06)
    if isinstance(obj, TrxFile):
        obj.close()


@pytest.mark.parametrize("path", [("gs.trk"), ("gs.trx"), ("gs_fldr.trx")])
@pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
def test_multi_load_save_rasmm(path):
    with TemporaryDirectory() as tmp_gs_dir:
        gs_dir = os.path.join(get_home(), "gold_standard")
        basename, ext = os.path.splitext(path)

        path = os.path.join(gs_dir, path)
        coord = np.loadtxt(
            os.path.join(get_home(), "gold_standard", "gs_rasmm_space.txt")
        )

        obj = load(path, os.path.join(gs_dir, "gs.nii"))
        for i in range(3):
            out_path = os.path.join(tmp_gs_dir, "{}_tmp{}_{}".format(basename, i, ext))
            save(obj, out_path)

            if isinstance(obj, TrxFile):
                obj.close()
            obj = load(out_path, os.path.join(gs_dir, "gs.nii"))

        assert_allclose(obj.streamlines._data, coord, rtol=1e-04, atol=1e-06)
        if isinstance(obj, TrxFile):
            obj.close()


@pytest.mark.parametrize("path", [("gs.trx"), ("gs_fldr.trx")])
@pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
def test_delete_tmp_gs_dir(path):
    gs_dir = os.path.join(get_home(), "gold_standard")
    path = os.path.join(gs_dir, path)

    trx1 = tmm.load(path)
    if os.path.isfile(path):
        tmp_gs_dir = deepcopy(trx1._uncompressed_folder_handle.name)
        assert os.path.isdir(tmp_gs_dir)
    sft = trx1.to_sft()
    trx1.close()

    coord_rasmm = np.loadtxt(
        os.path.join(get_home(), "gold_standard", "gs_rasmm_space.txt")
    )
    coord_vox = np.loadtxt(
        os.path.join(get_home(), "gold_standard", "gs_vox_space.txt")
    )

    # The folder trx representation does not need tmp files
    if os.path.isfile(path):
        assert not os.path.isdir(tmp_gs_dir)

    assert_allclose(sft.streamlines._data, coord_rasmm, rtol=1e-04, atol=1e-06)

    # Reloading the TRX and checking its data, then closing
    trx2 = tmm.load(path)
    assert_allclose(
        trx2.streamlines._data, sft.streamlines._data, rtol=1e-04, atol=1e-06
    )
    trx2.close()

    sft.to_vox()
    assert_allclose(sft.streamlines._data, coord_vox, rtol=1e-04, atol=1e-06)

    trx3 = tmm.load(path)
    assert_allclose(trx3.streamlines._data, coord_rasmm, rtol=1e-04, atol=1e-06)
    trx3.close()


@pytest.mark.parametrize("path", [("gs.trx")])
@pytest.mark.skipif(not dipy_available, reason="Dipy is not installed.")
def test_close_tmp_files(path):
    gs_dir = os.path.join(get_home(), "gold_standard")
    path = os.path.join(gs_dir, path)

    trx = tmm.load(path)
    process = psutil.Process(os.getpid())
    open_files = process.open_files()

    expected_content = [
        "offsets.uint32",
        "positions.3.float32",
        "header.json",
        "random_coord.3.float32",
        "color_y.float32",
        "color_x.float32",
        "color_z.float32",
    ]

    count = 0
    for open_file in open_files:
        basename = os.path.basename(open_file.path)
        if basename in expected_content:
            count += 1

    assert count == 6
    trx.close()

    open_files = process.open_files()
    count = 0
    for open_file in open_files:
        basename = os.path.basename(open_file.path)
        if basename in expected_content:
            count += 1
    assert not count


@pytest.mark.parametrize("tmp_path", [("~"), ("use_working_dir")])
def test_change_tmp_dir(tmp_path):
    gs_dir = os.path.join(get_home(), "gold_standard")
    path = os.path.join(gs_dir, "gs.trx")

    if tmp_path == "use_working_dir":
        os.environ["TRX_TMPDIR"] = "use_working_dir"
    else:
        os.environ["TRX_TMPDIR"] = os.path.expanduser(tmp_path)

    trx = tmm.load(path)
    tmp_gs_dir = deepcopy(trx._uncompressed_folder_handle.name)

    if tmp_path == "use_working_dir":
        assert os.path.dirname(tmp_gs_dir) == os.getcwd()
    else:
        assert os.path.dirname(tmp_gs_dir) == os.path.expanduser(tmp_path)

    trx.close()
    assert not os.path.isdir(tmp_gs_dir)


@pytest.mark.parametrize("path", [("gs.trx"), ("gs_fldr.trx")])
def test_complete_dir_from_trx(path):
    gs_dir = os.path.join(get_home(), "gold_standard")
    path = os.path.join(gs_dir, path)

    trx = tmm.load(path)
    if trx._uncompressed_folder_handle is None:
        dir_to_check = path
    else:
        dir_to_check = trx._uncompressed_folder_handle.name

    file_paths = []
    for dirpath, _, filenames in os.walk(dir_to_check):
        for filename in filenames:
            full_path = os.path.join(dirpath, filename)
            cut_path = full_path.split(dir_to_check)[1][1:].replace("\\", "/")
            file_paths.append(cut_path)

    expected_content = [
        "offsets.uint32",
        "positions.3.float32",
        "header.json",
        "dps/random_coord.3.float32",
        "dpv/color_y.float32",
        "dpv/color_x.float32",
        "dpv/color_z.float32",
    ]
    assert set(file_paths) == set(expected_content)


def test_complete_zip_from_trx():
    gs_dir = os.path.join(get_home(), "gold_standard")
    path = os.path.join(gs_dir, "gs.trx")

    with zipfile.ZipFile(path, mode="r") as zf:
        zip_file_list = zf.namelist()

    expected_content = [
        "offsets.uint32",
        "positions.3.float32",
        "header.json",
        "dps/random_coord.3.float32",
        "dpv/color_y.float32",
        "dpv/color_x.float32",
        "dpv/color_z.float32",
    ]
    assert set(zip_file_list) == set(expected_content)



"""TRX file format for brain tractography data."""

try:
    from ._version import __version__  # noqa: F401
except ImportError:
    try:
        from importlib.metadata import PackageNotFoundError, version

        __version__ = version("trx-python")
    except (ImportError, PackageNotFoundError):
        __version__ = "unknown"



# -*- coding: utf-8 -*-

import json
import os
import struct
import tempfile
import zipfile

from nibabel.streamlines import LazyTractogram
from nibabel.streamlines.tests.test_tractogram import make_dummy_streamline
import numpy as np
import pytest

try:
    import dipy  # noqa: F401

    dipy_available = True
except ImportError:
    dipy_available = False

from trx.fetcher import fetch_data, get_home, get_testing_files_dict
from trx.io import get_trx_tmp_dir
import trx.trx_file_memmap as tmm

fetch_data(get_testing_files_dict(), keys=["memmap_test_data.zip"])
tmp_dir = get_trx_tmp_dir()


@pytest.mark.parametrize(
    "arr,expected,value_error",
    [
        (np.ones((5, 5, 5), dtype=np.int16), None, True),
        (np.ones((5, 4), dtype=np.int16), "mean_fa.4.int16", False),
        (np.ones((5, 4), dtype=np.float64), "mean_fa.4.float64", False),
        (np.ones((5, 1), dtype=np.float64), "mean_fa.float64", False),
        (np.ones((1), dtype=np.float64), "mean_fa.float64", False),
    ],
)
def test__generate_filename_from_data(
    arr, expected, value_error, filename="mean_fa.bit"
):
    if value_error:
        with pytest.raises(ValueError):
            new_fn = tmm._generate_filename_from_data(arr=arr, filename=filename)
            assert new_fn is None
    else:
        new_fn = tmm._generate_filename_from_data(arr=arr, filename=filename)
        assert new_fn == expected


@pytest.mark.parametrize(
    "filename,expected,value_error",
    [
        ("mean_fa.float64", ("mean_fa", 1, ".float64"), False),
        ("mean_fa.5.int32", ("mean_fa", 5, ".int32"), False),
        ("mean_fa", None, True),
        ("mean_fa.5.4.int32", None, True),
        pytest.param(
            "mean_fa.fa", None, True, marks=pytest.mark.xfail, id="invalid extension"
        ),
    ],
)
def test__split_ext_with_dimensionality(filename, expected, value_error):
    if value_error:
        with pytest.raises(ValueError):
            assert tmm._split_ext_with_dimensionality(filename) == expected
    else:
        assert tmm._split_ext_with_dimensionality(filename) == expected


@pytest.mark.parametrize(
    "offsets,nb_vertices,expected",
    [
        (np.array(range(5), dtype=np.int16), 4, np.array([1, 1, 1, 1, 0])),
        (np.array([0, 1, 1, 3, 4], dtype=np.int32), 4, np.array([1, 0, 2, 1, 0])),
        (np.array(range(4), dtype=np.uint64), 4, np.array([1, 1, 1, 1])),
        pytest.param(
            np.array([0, 1, 0, 3, 4], dtype=np.int16),
            4,
            np.array([1, 3, 0, 1, 0]),
            marks=pytest.mark.xfail,
            id="offsets not sorted",
        ),
    ],
)
def test__compute_lengths(offsets, nb_vertices, expected):
    offsets = tmm._append_last_offsets(offsets, nb_vertices)
    lengths = tmm._compute_lengths(offsets=offsets)
    assert np.array_equal(lengths, expected)


@pytest.mark.parametrize(
    "ext,expected",
    [
        (".bit", True),
        (".int16", True),
        (".float32", True),
        (".ushort", True),
        (".txt", False),
    ],
)
def test__is_dtype_valid(ext, expected):
    assert tmm._is_dtype_valid(ext) == expected


@pytest.mark.parametrize(
    "arr,l_bound,r_bound,expected",
    [
        (np.array(range(5), dtype=np.int16), None, None, 4),
        (np.array([0, 1, 0, 3, 4], dtype=np.int16), None, None, 1),
        (np.array([0, 1, 2, 0, 4], dtype=np.int16), None, None, 2),
        (np.array(range(5), dtype=np.int16), 1, 2, 2),
        (np.array(range(5), dtype=np.int16), 3, 3, 3),
        (np.zeros((5), dtype=np.int16), 3, 3, -1),
    ],
)
def test__dichotomic_search(arr, l_bound, r_bound, expected):
    end_idx = tmm._dichotomic_search(arr, l_bound=l_bound, r_bound=r_bound)
    assert end_idx == expected


@pytest.mark.parametrize(
    "basename, create, expected",
    [
        ("offsets.int16", True, np.array(range(12), dtype=np.int16).reshape((3, 4))),
        ("offsets.float32", False, None),
    ],
)
def test__create_memmap(basename, create, expected):
    if create:
        with get_trx_tmp_dir() as dirname:
            filename = os.path.join(dirname, basename)
            fp = tmm._create_memmap(
                filename=filename, mode="w+", shape=(3, 4), dtype=np.int16
            )
            fp[:] = expected[:]
            mmarr = tmm._create_memmap(filename=filename, shape=(3, 4), dtype=np.int16)
            assert np.array_equal(mmarr, expected)

    else:
        with get_trx_tmp_dir() as dirname:
            filename = os.path.join(dirname, basename)
            mmarr = tmm._create_memmap(filename=filename, shape=(0,), dtype=np.int16)
            assert os.path.isfile(filename)
            assert np.array_equal(mmarr, np.zeros(shape=(0,), dtype=np.float32))


# need dpg test with missing keys
@pytest.mark.parametrize(
    "path,check_dpg,value_error",
    [
        ("small_compressed.trx", False, False),
        ("small.trx", True, False),
        ("small_fldr.trx", False, False),
        ("dontexist.trx", False, True),
    ],
)
def test_load(path, check_dpg, value_error):
    path = os.path.join(get_home(), "memmap_test_data", path)
    # Need to perhaps improve test
    if value_error:
        with pytest.raises(ValueError):
            assert not isinstance(
                tmm.load(input_obj=path, check_dpg=check_dpg), tmm.TrxFile
            )
    else:
        assert isinstance(tmm.load(input_obj=path, check_dpg=check_dpg), tmm.TrxFile)


@pytest.mark.parametrize("path", [("small.trx")])
def test_load_zip(path):
    path = os.path.join(get_home(), "memmap_test_data", path)
    assert isinstance(tmm.load_from_zip(path), tmm.TrxFile)


@pytest.mark.parametrize("path", [("small_fldr.trx")])
def test_load_directory(path):
    path = os.path.join(get_home(), "memmap_test_data", path)
    assert isinstance(tmm.load_from_directory(path), tmm.TrxFile)


@pytest.mark.parametrize("path", [("small.trx")])
def test_concatenate(path):
    path = os.path.join(get_home(), "memmap_test_data", path)
    trx1 = tmm.load(path)
    trx2 = tmm.load(path)
    concat = tmm.concatenate([trx1, trx2])

    assert len(concat) == 2 * len(trx2)
    trx1.close()
    trx2.close()
    concat.close()


@pytest.mark.parametrize("path", [("small.trx")])
def test_resize(path):
    path = os.path.join(get_home(), "memmap_test_data", path)
    trx1 = tmm.load(path)
    concat = tmm.TrxFile(nb_vertices=1000000, nb_streamlines=10000, init_as=trx1)

    tmm.concatenate([concat, trx1], preallocation=True, delete_groups=True)
    concat.resize()

    assert len(concat) == len(trx1)
    trx1.close()
    concat.close()


@pytest.mark.parametrize("path, buffer", [("small.trx", 10000), ("small.trx", 0)])
def test_append(path, buffer):
    path = os.path.join(get_home(), "memmap_test_data", path)
    trx1 = tmm.load(path)
    concat = tmm.TrxFile(nb_vertices=1, nb_streamlines=1, init_as=trx1)

    concat.append(trx1, extra_buffer=buffer)
    if buffer > 0:
        concat.resize()

    assert len(concat) == len(trx1)
    trx1.close()
    concat.close()


@pytest.mark.parametrize("path, buffer", [("small.trx", 10000)])
@pytest.mark.skipif(not dipy_available, reason="Dipy is not installed")
def test_append_StatefulTractogram(path, buffer):
    path = os.path.join(get_home(), "memmap_test_data", path)
    trx = tmm.load(path)
    obj = trx.to_sft()
    concat = tmm.TrxFile(nb_vertices=1, nb_streamlines=1, init_as=trx)

    concat.append(obj, extra_buffer=buffer)
    if buffer > 0:
        concat.resize()

    assert len(concat) == len(obj)
    trx.close()
    concat.close()


@pytest.mark.parametrize("path, buffer", [("small.trx", 10000)])
def test_append_Tractogram(path, buffer):
    path = os.path.join(get_home(), "memmap_test_data", path)
    trx = tmm.load(path)
    obj = trx.to_tractogram()
    concat = tmm.TrxFile(nb_vertices=1, nb_streamlines=1, init_as=trx)

    concat.append(obj, extra_buffer=buffer)
    if buffer > 0:
        concat.resize()

    assert len(concat) == len(obj)
    trx.close()
    concat.close()


@pytest.mark.parametrize(
    "path, size, buffer",
    [
        ("small.trx", 50, 10000),
        ("small.trx", 0, 10000),
        ("small.trx", 25000, 10000),
        ("small.trx", 50, 0),
        ("small.trx", 0, 0),
        ("small.trx", 25000, 10000),
    ],
)
def test_from_lazy_tractogram(path, size, buffer):
    _ = np.random.RandomState(1776)
    streamlines = []
    fa = []
    commit_weights = []
    clusters_QB = []
    gen_range = [1, 2, 5, 2, 1] * (size // 5)
    for i in gen_range:
        data = make_dummy_streamline(i)
        streamline, data_per_point, data_for_streamline = data
        streamlines.append(streamline)
        fa.append(data_per_point["fa"].astype(np.float16))
        commit_weights.append(data_for_streamline["mean_curvature"].astype(np.float32))
        clusters_QB.append(data_for_streamline["mean_torsion"].astype(np.uint16))

    def streamlines_func():
        return (e for e in streamlines)

    data_per_point_func = {"fa": lambda: (e for e in fa)}
    data_per_streamline_func = {
        "commit_weights": lambda: (e for e in commit_weights),
        "clusters_QB": lambda: (e for e in clusters_QB),
    }

    obj = LazyTractogram(
        streamlines_func,
        data_per_streamline_func,
        data_per_point_func,
        affine_to_rasmm=np.eye(4),
    )

    dtype_dict = {
        "positions": np.float32,
        "offsets": np.uint32,
        "dpv": {"fa": np.float16},
        "dps": {"commit_weights": np.float32, "clusters_QB": np.uint16},
    }
    path = os.path.join(get_home(), "memmap_test_data", path)
    trx = tmm.TrxFile.from_lazy_tractogram(
        obj, reference=path, extra_buffer=buffer, chunk_size=1000, dtype_dict=dtype_dict
    )

    assert len(trx) == len(gen_range)


def test_zip_from_folder():
    pass


def test_trxfile_init():
    pass


def test_trxfile_print():
    pass


def test_trxfile_len():
    fake = tmm.TrxFile(nb_vertices=100, nb_streamlines=10)
    assert len(fake) == 10


def test_trxfile_getitem():
    pass


def test_trxfile_deepcopy():
    pass


def test_get_real_len():
    fake = tmm.TrxFile(nb_vertices=100, nb_streamlines=10)
    assert fake._get_real_len() == (0, 0)


def test_copy_fixed_arrays_from():
    pass


def test_initialize_empty_trx():
    """Test creating, saving, and loading an empty TRX file."""
    trx = tmm.TrxFile()
    assert trx.header["NB_STREAMLINES"] == 0
    assert trx.header["NB_VERTICES"] == 0
    assert len(trx.streamlines) == 0

    with tempfile.TemporaryDirectory() as tmp_dir:
        out_path = os.path.join(tmp_dir, "empty.trx")
        tmm.save(trx, out_path)

        assert os.path.exists(out_path)
        file_size = os.path.getsize(out_path)
        assert file_size < 500  # Should be very small, just header.json in zip

        with zipfile.ZipFile(out_path, "r") as zf:
            filenames = [info.filename for info in zf.filelist]
            assert "header.json" in filenames
            positions_files = [f for f in filenames if f.startswith("positions")]
            offsets_files = [f for f in filenames if f.startswith("offsets")]
            assert len(positions_files) == 0
            assert len(offsets_files) == 0

        loaded_trx = tmm.load(out_path)
        assert loaded_trx.header["NB_STREAMLINES"] == 0
        assert loaded_trx.header["NB_VERTICES"] == 0
        assert len(loaded_trx.streamlines) == 0
        assert len(loaded_trx.groups) == 0
        assert len(loaded_trx.data_per_streamline) == 0
        assert len(loaded_trx.data_per_vertex) == 0
        assert len(loaded_trx.data_per_group) == 0
        loaded_trx.close()


def test_create_trx_from_pointer():
    pass


def test_trxfile_getgroup():
    pass


def test_trxfile_select():
    pass


def test_trxfile_to_memory():
    pass


def test_trxfile_close():
    pass


@pytest.mark.parametrize("path", [("small.trx")])
def test_close_releases_mmap_from_zip(path):
    """close() must release mmap handles even when loaded via load_from_zip()."""
    path = os.path.join(get_home(), "memmap_test_data", path)
    trx = tmm.load_from_zip(path)

    assert trx._uncompressed_folder_handle is None

    mmap_obj = trx.streamlines._data._mmap
    assert mmap_obj is not None, "expected a live mmap before close()"
    assert not mmap_obj.closed, "mmap should be open before close()"

    trx.close()

    assert mmap_obj.closed, (
        "mmap is still open after close() — the mmap teardown was skipped "
        "because _uncompressed_folder_handle was None"
    )


# Endianness tests for cross-platform compatibility (Issue #83)
@pytest.mark.parametrize(
    "dtype_input,expected_byteorder",
    [
        # Native dtypes should be converted to little-endian
        (np.float32, "<"),
        (np.float64, "<"),
        (np.int32, "<"),
        (np.int64, "<"),
        (np.uint32, "<"),
        (np.uint64, "<"),
        ("float32", "<"),
        ("float64", "<"),
        # Big-endian dtypes should be converted to little-endian
        (">f4", "<"),
        (">f8", "<"),
        (">i4", "<"),
        (">u4", "<"),
        # Little-endian dtypes should remain little-endian
        ("<f4", "<"),
        ("<i4", "<"),
        # Single-byte types don't have endianness (byteorder is '|')
        (np.uint8, "|"),
        (np.int8, "|"),
        (np.bool_, "|"),
    ],
)
def test__get_dtype_little_endian(dtype_input, expected_byteorder):
    """Test that _get_dtype_little_endian correctly converts dtypes."""
    result = tmm._get_dtype_little_endian(dtype_input)
    assert result.byteorder == expected_byteorder or (
        result.byteorder == "=" and np.little_endian and expected_byteorder == "<"
    )


@pytest.mark.parametrize(
    "dtype,test_value",
    [
        (np.float32, 3.14159),
        (np.float64, 2.71828),
        (np.int32, 12345),
        (np.int64, 9876543210),
        (np.uint32, 0xDEADBEEF),
        (np.uint64, 0xDEADBEEFCAFEBABE),
    ],
)
def test__ensure_little_endian(dtype, test_value):
    """Test that _ensure_little_endian correctly converts arrays."""
    # Create array in native byte order
    arr = np.array([test_value], dtype=dtype)

    # Ensure little endian
    result = tmm._ensure_little_endian(arr)

    # Result should be little-endian (or native if system is little-endian)
    assert result.dtype.byteorder in ("<", "=", "|")

    # Values should be preserved (use isclose for float types due to precision)
    if np.issubdtype(dtype, np.floating):
        assert np.isclose(result[0], test_value, rtol=1e-6)
    else:
        assert result[0] == test_value


def test__ensure_little_endian_big_endian_input():
    """Test _ensure_little_endian with explicitly big-endian input."""
    # Create a big-endian array
    big_endian_dtype = np.dtype(">u4")
    arr = np.array([0x12345678], dtype=big_endian_dtype)

    # Ensure little endian
    result = tmm._ensure_little_endian(arr)

    # Result should be little-endian
    assert result.dtype.byteorder == "<"

    # Value should be preserved
    assert result[0] == 0x12345678


def test_load_zip_with_local_header_extra_field():
    """Test loading ZIP where local header has extra field not in central dir.

    Regression test for a bug where zip_info.FileHeader() was used to calculate
    data offset. The ZIP spec allows local headers to have different extra
    fields than central directory entries. The fix reads the actual local
    file header to get the correct offset.
    """
    positions = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], dtype=np.float32)
    offsets = np.array([0, 2], dtype=np.uint64)
    header = {
        "DIMENSIONS": [10, 10, 10],
        "VOXEL_TO_RASMM": np.eye(4).tolist(),
        "NB_VERTICES": 2,
        "NB_STREAMLINES": 1,
    }

    with tempfile.TemporaryDirectory() as tmp_dir:
        trx_path = os.path.join(tmp_dir, "test.trx")

        # Build ZIP with extra bytes in local headers but not central directory
        with open(trx_path, "wb") as f:
            local_info = []
            extra = b"\x00\x00\x04\x00TEST"  # 8-byte extra field

            for name, data in [
                ("header.json", json.dumps(header).encode()),
                ("positions.3.float32", positions.tobytes()),
                ("offsets.uint64", offsets.tobytes()),
            ]:
                offset = f.tell()
                fname = name.encode()
                crc = zipfile.crc32(data)
                # Local header WITH extra field
                f.write(
                    struct.pack(
                        "<4sHHHHHIIIHH",
                        b"PK\x03\x04",
                        20,
                        0,
                        0,
                        0,
                        0,
                        crc,
                        len(data),
                        len(data),
                        len(fname),
                        len(extra),
                    )
                )
                f.write(fname)
                f.write(extra)
                f.write(data)
                local_info.append((name, offset, crc, len(data)))

            cd_start = f.tell()
            for name, offset, crc, size in local_info:
                fname = name.encode()
                # Central directory WITHOUT extra field (mismatch!)
                f.write(
                    struct.pack(
                        "<4sHHHHHHIIIHHHHHII",
                        b"PK\x01\x02",
                        20,
                        20,
                        0,
                        0,
                        0,
                        0,
                        crc,
                        size,
                        size,
                        len(fname),
                        0,
                        0,
                        0,
                        0,
                        0,
                        offset,
                    )
                )
                f.write(fname)

            # End of central directory
            f.write(
                struct.pack(
                    "<4sHHHHIIH",
                    b"PK\x05\x06",
                    0,
                    0,
                    3,
                    3,
                    f.tell() - cd_start,
                    cd_start,
                    0,
                )
            )

        trx = tmm.load_from_zip(trx_path)
        np.testing.assert_array_almost_equal(trx.streamlines._data, positions)
        assert trx.header["NB_VERTICES"] == 2
        assert trx.header["NB_STREAMLINES"] == 1

        trx.close()


def test_endianness_roundtrip():
    """Test that data survives write/read cycle with correct endianness."""
    with get_trx_tmp_dir() as dirname:
        # Test values that would be corrupted if endianness is wrong
        test_positions = np.array(
            [[1.5, 2.5, 3.5], [4.5, 5.5, 6.5], [7.5, 8.5, 9.5]], dtype=np.float32
        )
        test_offsets = np.array([0, 3], dtype=np.uint32)

        # Write as little-endian
        pos_file = os.path.join(dirname, "test_positions.3.float32")
        off_file = os.path.join(dirname, "test_offsets.uint32")

        tmm._ensure_little_endian(test_positions).tofile(pos_file)
        tmm._ensure_little_endian(test_offsets).tofile(off_file)

        # Read back using _create_memmap (which enforces little-endian)
        read_positions = tmm._create_memmap(
            pos_file, mode="r", shape=(3, 3), dtype="float32"
        )
        read_offsets = tmm._create_memmap(
            off_file, mode="r", shape=(2,), dtype="uint32"
        )

        # Values should match
        np.testing.assert_array_almost_equal(read_positions, test_positions)
        np.testing.assert_array_equal(read_offsets, test_offsets)



# -*- coding: utf-8 -*-
"""Test data management for downloading and verifying test assets."""

import hashlib
import logging
import os
import shutil
import urllib.request

TEST_DATA_REPO = "tee-ar-ex/trx-test-data"
TEST_DATA_TAG = "v0.1.0"
# GitHub release API entrypoint for metadata (asset list, sizes, etc.).
TEST_DATA_API_URL = (
    f"https://api.github.com/repos/{TEST_DATA_REPO}/releases/tags/{TEST_DATA_TAG}"
)
# Direct download base for release assets.
TEST_DATA_BASE_URL = (
    f"https://github.com/{TEST_DATA_REPO}/releases/download/{TEST_DATA_TAG}"
)


def get_home():
    """Return a user-writeable file-system location to put files.

    Returns
    -------
    str
        Path to the TRX home directory.
    """
    if "TRX_HOME" in os.environ:
        trx_home = os.environ["TRX_HOME"]
    else:
        trx_home = os.path.join(os.path.expanduser("~"), ".tee_ar_ex")
    return trx_home


def get_testing_files_dict():
    """Return dictionary linking zip file to their GitHub release URL and checksums.

    Assets are hosted under the v0.1.0 release of tee-ar-ex/trx-test-data.
    If URLs change, check TEST_DATA_API_URL to discover the latest asset
    locations.

    Returns
    -------
    dict
        Mapping of filenames to (url, md5, sha256) tuples.
    """
    return {
        "DSI.zip": (
            f"{TEST_DATA_BASE_URL}/DSI.zip",
            "b847f053fc694d55d935c0be0e5268f7",  # md5
            "1b09ce8b4b47b2600336c558fdba7051218296e8440e737364f2c4b8ebae666c",
        ),
        "memmap_test_data.zip": (
            f"{TEST_DATA_BASE_URL}/memmap_test_data.zip",
            "03f7651a0f9e3eeabee9aed0ad5f69e1",  # md5
            "98ba89d7a9a7baa2d37956a0a591dce9bb4581bd01296ad5a596706ee90a52ef",
        ),
        "trx_from_scratch.zip": (
            f"{TEST_DATA_BASE_URL}/trx_from_scratch.zip",
            "d9f220a095ce7f027772fcd9451a2ee5",  # md5
            "f98ab6da6a6065527fde4b0b6aa40f07583e925d952182e9bbd0febd55c0f6b2",
        ),
        "gold_standard.zip": (
            f"{TEST_DATA_BASE_URL}/gold_standard.zip",
            "57e3f9951fe77245684ede8688af3ae8",  # md5
            "35a0b633560cc2b0d8ecda885aa72d06385499e0cd1ca11a956b0904c3358f01",
        ),
    }


def md5sum(filename):
    """Compute the MD5 checksum of a file.

    Parameters
    ----------
    filename : str
        Path to file to hash.

    Returns
    -------
    str
        Hexadecimal MD5 digest.
    """
    h = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(128 * h.block_size), b""):
            h.update(chunk)
    return h.hexdigest()


def sha256sum(filename):
    """Compute the SHA256 checksum of a file.

    Parameters
    ----------
    filename : str
        Path to file to hash.

    Returns
    -------
    str
        Hexadecimal SHA256 digest.
    """
    h = hashlib.sha256()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(128 * h.block_size), b""):
            h.update(chunk)
    return h.hexdigest()


def fetch_data(files_dict, keys=None):  # noqa: C901
    """Download files to folder and check their md5 checksums.

    Parameters
    ----------
    files_dict : dict
        For each file in `files_dict` the value should be (url, md5).
        The file will be downloaded from url, if the file does not already
        exist or if the file exists but the md5 checksum does not match.
        Zip files are automatically unzipped and its contents are md5 checked.
    keys : list of str or str or None, optional
        Subset of keys from ``files_dict`` to download. When None, all
        keys are downloaded.

    Raises
    ------
    ValueError
        Raises if the md5 checksum of the file does not match the expected
        value. The downloaded file is not deleted when this error is raised.
    """
    trx_home = get_home()

    if not os.path.exists(trx_home):
        os.makedirs(trx_home)

    if keys is None:
        keys = files_dict.keys()
    elif isinstance(keys, str):
        keys = [keys]

    for f in keys:
        file_entry = files_dict[f]
        if len(file_entry) == 2:
            url, expected_md5 = file_entry
            expected_sha = None
        else:
            url, expected_md5, expected_sha = file_entry
        full_path = os.path.join(trx_home, f)

        logging.info("Downloading {} to {}".format(f, trx_home))
        if not os.path.exists(full_path):
            urllib.request.urlretrieve(url, full_path)

        actual_md5 = md5sum(full_path)
        if expected_md5 != actual_md5:
            raise ValueError(
                f"Md5sum for {f} does not match. "
                "Please remove the file to download it again: " + full_path
            )

        if expected_sha is not None:
            actual_sha = sha256sum(full_path)
            if expected_sha != actual_sha:
                raise ValueError(
                    f"SHA256 for {f} does not match. "
                    "Please remove the file to download it again: " + full_path
                )

        if f.endswith(".zip"):
            dst_dir = os.path.join(trx_home, f[:-4])
            shutil.unpack_archive(full_path, extract_dir=dst_dir, format="zip")



# -*- coding: utf-8 -*-

import numpy as np
import pytest

from trx.streamlines_ops import (
    difference,
    intersection,
    perform_streamlines_operation,
    union,
)

streamlines_ori = [
    np.ones(90).reshape((30, 3)),
    np.arange(90).reshape((30, 3)) + 0.3333,
]


@pytest.mark.parametrize(
    "precision, noise, expected",
    [
        (0, 0.0001, [4]),
        (1, 0.0001, [4]),
        (2, 0.0001, [4]),
        (4, 0.0001, []),
        (0, 0.01, [4]),
        (1, 0.01, [4]),
        (2, 0.01, []),
        (0, 1, []),
    ],
)
def test_intersection(precision, noise, expected):
    streamlines_new = []
    for i in range(5):
        if i < 4:
            streamlines_new.append(streamlines_ori[1] + np.random.random((30, 3)))
        else:
            streamlines_new.append(
                streamlines_ori[1] + noise * np.random.random((30, 3))
            )
    # print(streamlines_new)
    _, indices_uniq = perform_streamlines_operation(
        intersection, [streamlines_new, streamlines_ori], precision=precision
    )
    indices_uniq = indices_uniq.tolist()
    assert indices_uniq == expected


@pytest.mark.parametrize(
    "precision, noise, expected",
    [
        (0, 0.0001, 6),
        (1, 0.0001, 6),
        (2, 0.0001, 6),
        (4, 0.0001, 7),
        (0, 0.01, 6),
        (1, 0.01, 6),
        (2, 0.01, 7),
        (0, 1, 7),
    ],
)
def test_union(precision, noise, expected):
    streamlines_new = []
    for i in range(5):
        if i < 4:
            streamlines_new.append(streamlines_ori[1] + np.random.random((30, 3)))
        else:
            streamlines_new.append(
                streamlines_ori[1] + noise * np.random.random((30, 3))
            )

    unique_streamlines, _ = perform_streamlines_operation(
        union, [streamlines_new, streamlines_ori], precision=precision
    )
    assert len(unique_streamlines) == expected


@pytest.mark.parametrize(
    "precision, noise, expected",
    [
        (0, 0.0001, 4),
        (1, 0.0001, 4),
        (2, 0.0001, 4),
        (4, 0.0001, 5),
        (0, 0.01, 4),
        (1, 0.01, 4),
        (2, 0.01, 5),
        (0, 1, 5),
    ],
)
def test_difference(precision, noise, expected):
    streamlines_new = []
    for i in range(5):
        if i < 4:
            streamlines_new.append(streamlines_ori[1] + np.random.random((30, 3)))
        else:
            streamlines_new.append(
                streamlines_ori[1] + noise * np.random.random((30, 3))
            )

    unique_streamlines, _ = perform_streamlines_operation(
        difference, [streamlines_new, streamlines_ori], precision=precision
    )
    assert len(unique_streamlines) == expected
